import ipywidgets as widgets
from IPython.display import display
# from IPython.display import clear_output

from graphviz import Digraph
import copy
import os

from ..wrappers.database import sklearn_db, cheml_db, pandas_db
from ..wrappers.database.TSHF import tshf
from ..utils import isint

##########################################################
# Todo: bring back receivers - no need another recursive function for bidR; (currentbids - bidS = bidR)
# Todo: profile the cpu/clock time

class container_page(object):
    def __init__(self, title, index, widget, block_params={}):
        self.title = title
        self.index = index
        self.widget = widget
        self.block_params = block_params
        # IO_refresher will be added later for each function block

class wrapperGUI(object):
    def __init__(self):
        self.tasks, self.combinations = tshf()  # task, subtask, host, function connections; keep the tasks for the order
        self.blocks = {}  # similar to cmls in the parser, {'task':{}, 'subtask':{}, 'host':{}, 'function':{}, 'parameters':{}, 'send':{'NA':'here'}, 'recv':{'NA':'here'} }
        self.block_id = 1
        self.comp_graph = []    # list of (iblock_send,token,iblock_recv,token)
        self.out_dir = "CMLWrapper.out"
        self.pages = {}
        self.graph = widgets.Image()
        self.accordion = widgets.Accordion()
        self.home_page()

    ################################

    def display_accordion(self, active_id=0, layout=widgets.Layout(border='solid gray 2px')):
        def index_change(change):
            ind = self.accordion.selected_index
            if ind not in [0,1,None]:
                ib = [i for i in sorted(self.pages)][ind]
                self.pages[ib].IO_refresher()

        children = [self.pages[i].widget for i in sorted(self.pages)]
        self.accordion.children = children
        self.accordion.selected_index = active_id
        self.accordion.layout = layout
        new_titles = {}
        for i,ib in enumerate(sorted(self.pages)):
            if i in [0, 1]:
                self.accordion.set_title(i, '%s' % self.pages[i].title)
                # new_titles[str(i)] = '%s' % self.pages[i].title
            else:
                self.accordion.set_title(i, 'block# %i: %s' % (ib, self.pages[ib].title))
                # new_titles[str(i)] = 'block# %i: %s' % (ib, self.pages[ib].title)
        # self.accordion._titles = new_titles
        self.accordion.observe(index_change, names='selected_index')

    def close_accordion(self):
        pass
        # self.accordion.close()

    ################################

    def check_filename(self,filename):
        if filename is None:
            filename = 'cheml_config.txt'
        initial_file_name = filename
        i = 0
        while os.path.exists(filename):
            i+=1
            base = os.path.splitext(initial_file_name)[0]
            format = os.path.splitext(initial_file_name)[1]
            if format == '':
                format = 'txt'
            filename = base + '%i'%i + format
        return filename

    def configurer(self):
        def value(string):
            try:
                val =  eval(string)
                if isinstance(val,str):
                    val = '"' + val + '"'
                return val
            except NameError:
                return '"' + string + '"'

        tab = '    '
        lines = []
        uids = {}   #comp_graph : unique ids
        for i in range(len(self.comp_graph)):
            uids[self.comp_graph[i]] = i
        for ib in self.pages:
            if ib not in [0,1]:
                edges = [e for e in self.comp_graph if ib in e]
                task = self.pages[ib].block_params['task']
                subtask = self.pages[ib].block_params['subtask']
                if len(edges) == 0:
                    lines.append('# (%s,%s)\n'%(task,subtask))
                else:
                    lines.append('## (%s,%s)\n'%(task,subtask))
                host = self.pages[ib].block_params['host']
                lines.append(tab+"<< host = %s\n"%host)
                function = self.pages[ib].block_params['function']
                lines.append(tab+"<< function = %s\n"%function)
                for param in self.pages[ib].block_params['wparams']:
                    container = self.pages[ib].block_params['wparams'][param]
                    if container.checker.value:
                        lines.append(tab+"<< %s = %s\n"%(container.name, str(container.widget.value)))
                for param in self.pages[ib].block_params['fparams']:
                    container = self.pages[ib].block_params['fparams'][param]
                    if container.checker.value:
                        lines.append(tab+"<< %s = %s\n"%(container.name, str(container.widget.value)))
                for e in edges:
                    if e[0] == ib:
                        lines.append(tab+">> %s %i\n"%(e[1],uids[e]))
                    if e[2] == ib:
                        lines.append(tab + ">> %i %s\n" % (uids[e], e[3]))
                lines.append('\n')
        return lines

    def templates_widgets(self):
        headerT = widgets.Label(value='Start with a template workflow', layout=widgets.Layout(width='50%'))
        # style = {'description_width': 'initial'}
        # outdir = widgets.Text(
        #     value='CMLWrapper.out',
        #     placeholder='Type something',
        #     description='Output directory:',
        #     disabled=False,
        #     style = style,
        #     layout = widgets.Layout(margin='30px 0px 30px 0px'))21


        TEMPLATES = []

        #######################***********************#######################***********************
        #######################:::::::::::::::::::::::#######################
        ### Datasets Overview  (DATAOVER)
        headerDATAOVER = widgets.HTML(value='<b> Datasets Overview: </b>', layout=widgets.Layout(width='50%',margin='10px 0px 10px 0px'))
        TEMPLATES.append(headerDATAOVER)
        
        ######################################################
        ## Template7
        def on_selectTe7_clicked(b):
            # template7.txt is a cheml wrapper config file
            from .templates import template7
            script = template7()
            old = [i for i in self.pages]

            try:
                self.parser(script)
                # update the current_bid
                self.block_id = max(self.pages)
                selectTe7.icon = 'check'
            except Exception as err:
                print "Invalid configuration file ..."
                print "    IOError: %s"%err.message
                print "... Not loaded!"
                selectTe7.icon = 'remove'
                rm = [i for i in self.pages if i not in old]
                for ib in rm:
                    if ib in self.pages:
                        del self.pages[ib]

            self.debut = False
            self.add_page()

            ## clear ouput and update the graph viz
            self.graph.close()
            dot = Digraph(format='png')
            for edge in self.comp_graph:
                dot.node('%i' % edge[0], label='%i %s' % (edge[0], self.pages[edge[0]].title))
                dot.node('%i' % edge[2], label='%i %s' % (edge[2], self.pages[edge[2]].title))
                dot.edge('%i' % edge[0], '%i' % edge[2], label='%s > %s' % (edge[1], edge[3]), labelfontcolor='green')
            self.graph = widgets.Image(value=dot.pipe(), format='png')
            display(self.graph)

        te7 = widgets.Label(value="Template 1: load_cep_homo --> plot histogram of HOMO energies --> print 5 SMILES")#, layout=widgets.Layout(width='70%'))
        selectTe7 = widgets.Button(description="Select")
        selectTe7.style.button_color = 'lightblue'
        selectTe7.on_click(on_selectTe7_clicked)
        # viewT1 = widgets.Button(description="Overview")
        # viewT1.style.button_color = 'lightblue'
        # viewT1.on_click(on_viewT1_clicked)
        hboxTe7 = widgets.HBox([te7, selectTe7],layout=widgets.Layout( border='dotted black 1px',justify_content = 'space-between'))
                                                               # height='40px', align_items='center',   justify_content = 'space-between',
                                                               # margin='0px 0px 0px 10px'))
        TEMPLATES.append(hboxTe7)

        ######################################################
        ## Template8
        def on_selectTe8_clicked(b):
            # template8.txt is a cheml wrapper config file
            from .templates import template8
            script = template8()
            old = [i for i in self.pages]

            try:
                self.parser(script)
                # update the current_bid
                self.block_id = max(self.pages)
                selectTe8.icon = 'check'
            except Exception as err:
                print "Invalid configuration file ..."
                print "    IOError: %s"%err.message
                print "... Not loaded!"
                selectTe8.icon = 'remove'
                rm = [i for i in self.pages if i not in old]
                for ib in rm:
                    if ib in self.pages:
                        del self.pages[ib]

            self.debut = False
            self.add_page()

            ## clear ouput and update the graph viz
            self.graph.close()
            dot = Digraph(format='png')
            for edge in self.comp_graph:
                dot.node('%i' % edge[0], label='%i %s' % (edge[0], self.pages[edge[0]].title))
                dot.node('%i' % edge[2], label='%i %s' % (edge[2], self.pages[edge[2]].title))
                dot.edge('%i' % edge[0], '%i' % edge[2], label='%s > %s' % (edge[1], edge[3]), labelfontcolor='green')
            self.graph = widgets.Image(value=dot.pipe(), format='png')
            display(self.graph)

        te8 = widgets.Label(value="Template 2: load_organic_density --> plot feature AMW vs. density --> print 5 SMILES")#, layout=widgets.Layout(width='70%'))
        selectTe8 = widgets.Button(description="Select")
        selectTe8.style.button_color = 'lightblue'
        selectTe8.on_click(on_selectTe8_clicked)
        # viewT1 = widgets.Button(description="Overview")
        # viewT1.style.button_color = 'lightblue'
        # viewT1.on_click(on_viewT1_clicked)
        hboxTe8 = widgets.HBox([te8, selectTe8],layout=widgets.Layout( border='dotted black 1px',justify_content = 'space-between'))
                                                               # height='40px', align_items='center',   justify_content = 'space-between',
                                                               # margin='0px 0px 0px 10px'))
        TEMPLATES.append(hboxTe8)

        ######################################################
        ## Template9
        def on_selectTe9_clicked(b):
            # template9.txt is a cheml wrapper config file
            from .templates import template9
            script = template9()
            old = [i for i in self.pages]

            try:
                self.parser(script)
                # update the current_bid
                self.block_id = max(self.pages)
                selectTe9.icon = 'check'
            except Exception as err:
                print "Invalid configuration file ..."
                print "    IOError: %s"%err.message
                print "... Not loaded!"
                selectTe9.icon = 'remove'
                rm = [i for i in self.pages if i not in old]
                for ib in rm:
                    if ib in self.pages:
                        del self.pages[ib]

            self.debut = False
            self.add_page()

            ## clear ouput and update the graph viz
            self.graph.close()
            dot = Digraph(format='png')
            for edge in self.comp_graph:
                dot.node('%i' % edge[0], label='%i %s' % (edge[0], self.pages[edge[0]].title))
                dot.node('%i' % edge[2], label='%i %s' % (edge[2], self.pages[edge[2]].title))
                dot.edge('%i' % edge[0], '%i' % edge[2], label='%s > %s' % (edge[1], edge[3]), labelfontcolor='green')
            self.graph = widgets.Image(value=dot.pipe(), format='png')
            display(self.graph)

        te9 = widgets.Label(value="Template 3: load_xyz_polarizability --> plot all 50 polarizabilities --> print first item of coordinates output")#, layout=widgets.Layout(width='70%'))
        selectTe9 = widgets.Button(description="Select")
        selectTe9.style.button_color = 'lightblue'
        selectTe9.on_click(on_selectTe9_clicked)
        # viewT1 = widgets.Button(description="Overview")
        # viewT1.style.button_color = 'lightblue'
        # viewT1.on_click(on_viewT1_clicked)
        hboxTe9 = widgets.HBox([te9, selectTe9],layout=widgets.Layout( border='dotted black 1px',justify_content = 'space-between'))
                                                               # height='40px', align_items='center',   justify_content = 'space-between',
                                                               # margin='0px 0px 0px 10px'))
        TEMPLATES.append(hboxTe9)



        #######################***********************#######################***********************
        #######################:::::::::::::::::::::::#######################:::::::::::::::::::::::
        ### Molecular Descriptors (MOLDES)
        headerMOLDES = widgets.HTML(value='<b> Molecular Descriptors: </b>', layout=widgets.Layout(width='50%',margin='10px 0px 10px 0px'))
        TEMPLATES.append(headerMOLDES)
        
        ######################################################
        ## Template1
        def on_selectTe1_clicked(b):
            # template1.txt is a cheml wrapper config file
            from .templates import template1
            script = template1()
            old = [i for i in self.pages]

            try:
                self.parser(script)
                # update the current_bid
                self.block_id = max(self.pages)
                selectTe1.icon = 'check'
            except Exception as err:
                print "Invalid configuration file ..."
                print "    IOError: %s"%err.message
                print "... Not loaded!"
                selectTe1.icon = 'remove'
                rm = [i for i in self.pages if i not in old]
                for ib in rm:
                    if ib in self.pages:
                        del self.pages[ib]

            self.debut = False
            self.add_page()

            ## clear ouput and update the graph viz
            self.graph.close()
            dot = Digraph(format='png')
            for edge in self.comp_graph:
                dot.node('%i' % edge[0], label='%i %s' % (edge[0], self.pages[edge[0]].title))
                dot.node('%i' % edge[2], label='%i %s' % (edge[2], self.pages[edge[2]].title))
                dot.edge('%i' % edge[0], '%i' % edge[2], label='%s > %s' % (edge[1], edge[3]), labelfontcolor='green')
            self.graph = widgets.Image(value=dot.pipe(), format='png')
            display(self.graph)

        te1 = widgets.Label(value="Template 1: load_xyz_polarizability --> generate CoulombMatrix features --> save features")#, layout=widgets.Layout(width='70%'))
        selectTe1 = widgets.Button(description="Select")
        selectTe1.style.button_color = 'lightblue'
        selectTe1.on_click(on_selectTe1_clicked)
        # viewT1 = widgets.Button(description="Overview")
        # viewT1.style.button_color = 'lightblue'
        # viewT1.on_click(on_viewT1_clicked)
        hboxTe1 = widgets.HBox([te1, selectTe1],layout=widgets.Layout( border='dotted black 1px',justify_content = 'space-between'))
                                                               # height='40px', align_items='center',   justify_content = 'space-between',
                                                               # margin='0px 0px 0px 10px'))
        TEMPLATES.append(hboxTe1)

        ######################################################
        ## Template2
        def on_selectTe2_clicked(b):
            # template1.txt is a cheml wrapper config file
            from .templates import template2
            script = template2()
            old = [i for i in self.pages]

            try:
                self.parser(script)
                # update the current_bid
                self.block_id = max(self.pages)
                selectTe2.icon = 'check'
            except Exception as err:
                print "Invalid configuration file ..."
                print "    IOError: %s"%err.message
                print "... Not loaded!"
                selectTe2.icon = 'remove'
                rm = [i for i in self.pages if i not in old]
                for ib in rm:
                    if ib in self.pages:
                        del self.pages[ib]

            self.debut = False
            self.add_page()

            ## clear ouput and update the graph viz
            self.graph.close()
            dot = Digraph(format='png')
            for edge in self.comp_graph:
                dot.node('%i' % edge[0], label='%i %s' % (edge[0], self.pages[edge[0]].title))
                dot.node('%i' % edge[2], label='%i %s' % (edge[2], self.pages[edge[2]].title))
                dot.edge('%i' % edge[0], '%i' % edge[2], label='%s > %s' % (edge[1], edge[3]), labelfontcolor='green')
            self.graph = widgets.Image(value=dot.pipe(), format='png')
            display(self.graph)

        te2 = widgets.Label(value="Template 2: load_xyz_polarizability --> generate BagofBonds features --> save features")#, layout=widgets.Layout(width='70%'))
        selectTe2 = widgets.Button(description="Select")
        selectTe2.style.button_color = 'lightblue'
        selectTe2.on_click(on_selectTe2_clicked)
        # viewT1 = widgets.Button(description="Overview")
        # viewT1.style.button_color = 'lightblue'
        # viewT1.on_click(on_viewT1_clicked)
        hboxTe2 = widgets.HBox([te2, selectTe2],layout=widgets.Layout( border='dotted black 1px',justify_content = 'space-between'))
                                                               # height='40px', align_items='center',   justify_content = 'space-between',
                                                               # margin='0px 0px 0px 10px'))
        TEMPLATES.append(hboxTe2)

        ######################################################
        ## Template3
        def on_selectTe3_clicked(b):
            # template3.txt is a cheml wrapper config file
            from .templates import template3
            script = template3()
            old = [i for i in self.pages]

            try:
                self.parser(script)
                # update the current_bid
                self.block_id = max(self.pages)
                selectTe3.icon = 'check'
            except Exception as err:
                print "Invalid configuration file ..."
                print "    IOError: %s"%err.message
                print "... Not loaded!"
                selectTe3.icon = 'remove'
                rm = [i for i in self.pages if i not in old]
                for ib in rm:
                    if ib in self.pages:
                        del self.pages[ib]

            self.debut = False
            self.add_page()

            ## clear ouput and update the graph viz
            self.graph.close()
            dot = Digraph(format='png')
            for edge in self.comp_graph:
                dot.node('%i' % edge[0], label='%i %s' % (edge[0], self.pages[edge[0]].title))
                dot.node('%i' % edge[2], label='%i %s' % (edge[2], self.pages[edge[2]].title))
                dot.edge('%i' % edge[0], '%i' % edge[2], label='%s > %s' % (edge[1], edge[3]), labelfontcolor='green')
            self.graph = widgets.Image(value=dot.pipe(), format='png')
            display(self.graph)

        te3 = widgets.Label(value="Template 3: get SMILES reperesentaion of molecules --> save them --> generate Morgan Fingerprints  --> save features")#, layout=widgets.Layout(width='70%'))
        selectTe3 = widgets.Button(description="Select")
        selectTe3.style.button_color = 'lightblue'
        selectTe3.on_click(on_selectTe3_clicked)
        # viewT1 = widgets.Button(description="Overview")
        # viewT1.style.button_color = 'lightblue'
        # viewT1.on_click(on_viewT1_clicked)
        hboxTe3 = widgets.HBox([te3, selectTe3],layout=widgets.Layout( border='dotted black 1px',justify_content = 'space-between'))
                                                               # height='40px', align_items='center',   justify_content = 'space-between',
                                                               # margin='0px 0px 0px 10px'))
        TEMPLATES.append(hboxTe3)

        ######################################################
        ## Template4
        def on_selectTe4_clicked(b):
            # template4.txt is a cheml wrapper config file
            from .templates import template4
            script = template4()
            old = [i for i in self.pages]

            try:
                self.parser(script)
                # update the current_bid
                self.block_id = max(self.pages)
                selectTe4.icon = 'check'
            except Exception as err:
                print "Invalid configuration file ..."
                print "    IOError: %s"%err.message
                print "... Not loaded!"
                selectTe4.icon = 'remove'
                rm = [i for i in self.pages if i not in old]
                for ib in rm:
                    if ib in self.pages:
                        del self.pages[ib]

            self.debut = False
            self.add_page()

            ## clear ouput and update the graph viz
            self.graph.close()
            dot = Digraph(format='png')
            for edge in self.comp_graph:
                dot.node('%i' % edge[0], label='%i %s' % (edge[0], self.pages[edge[0]].title))
                dot.node('%i' % edge[2], label='%i %s' % (edge[2], self.pages[edge[2]].title))
                dot.edge('%i' % edge[0], '%i' % edge[2], label='%s > %s' % (edge[1], edge[3]), labelfontcolor='green')
            self.graph = widgets.Image(value=dot.pipe(), format='png')
            display(self.graph)

        te4 = widgets.Label(value="Template 4: get SMILES reperesentaion of molecules --> save them --> generate Dragon descriptors  --> save features")#, layout=widgets.Layout(width='70%'))
        selectTe4 = widgets.Button(description="Select")
        selectTe4.style.button_color = 'lightblue'
        selectTe4.on_click(on_selectTe4_clicked)
        # viewT1 = widgets.Button(description="Overview")
        # viewT1.style.button_color = 'lightblue'
        # viewT1.on_click(on_viewT1_clicked)
        hboxTe4 = widgets.HBox([te4, selectTe4],layout=widgets.Layout( border='dotted black 1px',justify_content = 'space-between'))
                                                               # height='40px', align_items='center',   justify_content = 'space-between',
                                                               # margin='0px 0px 0px 10px'))
        TEMPLATES.append(hboxTe4)


        #######################***********************#######################***********************
        #######################:::::::::::::::::::::::#######################:::::::::::::::::::::::
        ### Inorganic Descriptors (InorgDes)
        headerInorgDes = widgets.HTML(value='<b> Inorganic Descriptors: </b>', layout=widgets.Layout(width='50%',margin='10px 0px 10px 0px'))
        TEMPLATES.append(headerInorgDes)

        ######################################################
        ## Template5
        def on_selectTe5_clicked(b):
            # template5.txt is a cheml wrapper config file
            from .templates import template5
            script = template5()
            old = [i for i in self.pages]

            try:
                self.parser(script)
                # update the current_bid
                self.block_id = max(self.pages)
                selectTe5.icon = 'check'
            except Exception as err:
                print "Invalid configuration file ..."
                print "    IOError: %s"%err.message
                print "... Not loaded!"
                selectTe5.icon = 'remove'
                rm = [i for i in self.pages if i not in old]
                for ib in rm:
                    if ib in self.pages:
                        del self.pages[ib]

            self.debut = False
            self.add_page()

            ## clear ouput and update the graph viz
            self.graph.close()
            dot = Digraph(format='png')
            for edge in self.comp_graph:
                dot.node('%i' % edge[0], label='%i %s' % (edge[0], self.pages[edge[0]].title))
                dot.node('%i' % edge[2], label='%i %s' % (edge[2], self.pages[edge[2]].title))
                dot.edge('%i' % edge[0], '%i' % edge[2], label='%s > %s' % (edge[1], edge[3]), labelfontcolor='green')
            self.graph = widgets.Image(value=dot.pipe(), format='png')
            display(self.graph)

        te5 = widgets.Label(value="Template 1: load_comp_energy --> inorganic descriptors for composition entries --> concatenate and print shape")#, layout=widgets.Layout(width='70%'))
        selectTe5 = widgets.Button(description="Select")
        selectTe5.style.button_color = 'lightblue'
        selectTe5.on_click(on_selectTe5_clicked)
        # viewT1 = widgets.Button(description="Overview")
        # viewT1.style.button_color = 'lightblue'
        # viewT1.on_click(on_viewT1_clicked)
        hboxTe5 = widgets.HBox([te5, selectTe5],layout=widgets.Layout( border='dotted black 1px',justify_content = 'space-between'))
                                                               # height='40px', align_items='center',   justify_content = 'space-between',
                                                               # margin='0px 0px 0px 10px'))
        TEMPLATES.append(hboxTe5)

        ######################################################
        ## Template6
        def on_selectTe6_clicked(b):
            # template6.txt is a cheml wrapper config file
            from .templates import template6
            script = template6()
            old = [i for i in self.pages]

            try:
                self.parser(script)
                # update the current_bid
                self.block_id = max(self.pages)
                selectTe6.icon = 'check'
            except Exception as err:
                print "Invalid configuration file ..."
                print "    IOError: %s"%err.message
                print "... Not loaded!"
                selectTe6.icon = 'remove'
                rm = [i for i in self.pages if i not in old]
                for ib in rm:
                    if ib in self.pages:
                        del self.pages[ib]

            self.debut = False
            self.add_page()

            ## clear ouput and update the graph viz
            self.graph.close()
            dot = Digraph(format='png')
            for edge in self.comp_graph:
                dot.node('%i' % edge[0], label='%i %s' % (edge[0], self.pages[edge[0]].title))
                dot.node('%i' % edge[2], label='%i %s' % (edge[2], self.pages[edge[2]].title))
                dot.edge('%i' % edge[0], '%i' % edge[2], label='%s > %s' % (edge[1], edge[3]), labelfontcolor='green')
            self.graph = widgets.Image(value=dot.pipe(), format='png')
            display(self.graph)

        te6 = widgets.Label(value="Template 2: load_crystal_structures --> inorganic descriptors for crystal structures --> concatenate and print shape")#, layout=widgets.Layout(width='70%'))
        selectTe6 = widgets.Button(description="Select")
        selectTe6.style.button_color = 'lightblue'
        selectTe6.on_click(on_selectTe6_clicked)
        # viewT1 = widgets.Button(description="Overview")
        # viewT1.style.button_color = 'lightblue'
        # viewT1.on_click(on_viewT1_clicked)
        hboxTe6 = widgets.HBox([te6, selectTe6],layout=widgets.Layout( border='dotted black 1px',justify_content = 'space-between'))
                                                               # height='40px', align_items='center',   justify_content = 'space-between',
                                                               # margin='0px 0px 0px 10px'))
        TEMPLATES.append(hboxTe6)


        #######################***********************#######################***********************
        #######################:::::::::::::::::::::::#######################:::::::::::::::::::::::
        ### Data Mining (DMine)
        headerDMine = widgets.HTML(value='<b> Data Mining: </b>', layout=widgets.Layout(width='50%',margin='10px 0px 10px 0px'))
        TEMPLATES.append(headerDMine)

        ######################################################
        ## Template11
        def on_selectTe11_clicked(b):
            # template11.txt is a cheml wrapper config file
            from .templates import template11
            script = template11()
            old = [i for i in self.pages]

            try:
                self.parser(script)
                # update the current_bid
                self.block_id = max(self.pages)
                selectTe11.icon = 'check'
            except Exception as err:
                print "Invalid configuration file ..."
                print "    IOError: %s"%err.message
                print "... Not loaded!"
                selectTe11.icon = 'remove'
                rm = [i for i in self.pages if i not in old]
                for ib in rm:
                    if ib in self.pages:
                        del self.pages[ib]

            self.debut = False
            self.add_page()

            ## clear ouput and update the graph viz
            self.graph.close()
            dot = Digraph(format='png')
            for edge in self.comp_graph:
                dot.node('%i' % edge[0], label='%i %s' % (edge[0], self.pages[edge[0]].title))
                dot.node('%i' % edge[2], label='%i %s' % (edge[2], self.pages[edge[2]].title))
                dot.edge('%i' % edge[0], '%i' % edge[2], label='%s > %s' % (edge[1], edge[3]), labelfontcolor='green')
            self.graph = widgets.Image(value=dot.pipe(), format='png')
            display(self.graph)

        te11 = widgets.Label(value="Template 1: model selection with grid search and cross validation")#, layout=widgets.Layout(width='70%'))
        selectTe11 = widgets.Button(description="Select")
        selectTe11.style.button_color = 'lightblue'
        selectTe11.on_click(on_selectTe11_clicked)
        # viewT1 = widgets.Button(description="Overview")
        # viewT1.style.button_color = 'lightblue'
        # viewT1.on_click(on_viewT1_clicked)
        hboxTe11 = widgets.HBox([te11, selectTe11],layout=widgets.Layout( border='dotted black 1px',justify_content = 'space-between'))
                                                               # height='40px', align_items='center',   justify_content = 'space-between',
                                                               # margin='0px 0px 0px 10px'))
        TEMPLATES.append(hboxTe11)

        ######################################################
        ## Template12
        def on_selectTe12_clicked(b):
            # template12.txt is a cheml wrapper config file
            from .templates import template12
            script = template12()
            old = [i for i in self.pages]

            try:
                self.parser(script)
                # update the current_bid
                self.block_id = max(self.pages)
                selectTe12.icon = 'check'
            except Exception as err:
                print "Invalid configuration file ..."
                print "    IOError: %s"%err.message
                print "... Not loaded!"
                selectTe12.icon = 'remove'
                rm = [i for i in self.pages if i not in old]
                for ib in rm:
                    if ib in self.pages:
                        del self.pages[ib]

            self.debut = False
            self.add_page()

            ## clear ouput and update the graph viz
            self.graph.close()
            dot = Digraph(format='png')
            for edge in self.comp_graph:
                dot.node('%i' % edge[0], label='%i %s' % (edge[0], self.pages[edge[0]].title))
                dot.node('%i' % edge[2], label='%i %s' % (edge[2], self.pages[edge[2]].title))
                dot.edge('%i' % edge[0], '%i' % edge[2], label='%s > %s' % (edge[1], edge[3]), labelfontcolor='green')
            self.graph = widgets.Image(value=dot.pipe(), format='png')
            display(self.graph)

        te12 = widgets.Label(value="Template 2: a complete machine learning workflow")#, layout=widgets.Layout(width='70%'))
        selectTe12 = widgets.Button(description="Select")
        selectTe12.style.button_color = 'lightblue'
        selectTe12.on_click(on_selectTe12_clicked)
        # viewT1 = widgets.Button(description="Overview")
        # viewT1.style.button_color = 'lightblue'
        # viewT1.on_click(on_viewT1_clicked)
        hboxTe12 = widgets.HBox([te12, selectTe12],layout=widgets.Layout( border='dotted black 1px',justify_content = 'space-between'))
                                                               # height='40px', align_items='center',   justify_content = 'space-between',
                                                               # margin='0px 0px 0px 10px'))
        TEMPLATES.append(hboxTe12)

        ######################################################
        # ## TemplateMHL
        # def on_selectTeMHL_clicked(b):
        #     # templateMHL.txt is a cheml wrapper config file
        #     from .templates import templateMHL
        #     script = templateMHL()
        #     old = [i for i in self.pages]
        #
        #     try:
        #         self.parser(script)
        #         # update the current_bid
        #         self.block_id = max(self.pages)
        #         selectTeMHL.icon = 'check'
        #     except Exception as err:
        #         print "Invalid configuration file ..."
        #         print "    IOError: %s"%err.message
        #         print "... Not loaded!"
        #         selectTeMHL.icon = 'remove'
        #         rm = [i for i in self.pages if i not in old]
        #         for ib in rm:
        #             if ib in self.pages:
        #                 del self.pages[ib]
        #
        #     self.debut = False
        #     self.add_page()
        #
        #     ## clear ouput and update the graph viz
        #     self.graph.close()
        #     dot = Digraph(format='png')
        #     for edge in self.comp_graph:
        #         dot.node('%i' % edge[0], label='%i %s' % (edge[0], self.pages[edge[0]].title))
        #         dot.node('%i' % edge[2], label='%i %s' % (edge[2], self.pages[edge[2]].title))
        #         dot.edge('%i' % edge[0], '%i' % edge[2], label='%s > %s' % (edge[1], edge[3]), labelfontcolor='green')
        #     self.graph = widgets.Image(value=dot.pipe(), format='png')
        #     display(self.graph)
        #
        # teMHL = widgets.Label(value="Template MHL: read XYZ files --> generate BagofBonds features --> save features")#, layout=widgets.Layout(width='70%'))
        # selectTeMHL = widgets.Button(description="Select")
        # selectTeMHL.style.button_color = 'lightblue'
        # selectTeMHL.on_click(on_selectTeMHL_clicked)
        # # viewT1 = widgets.Button(description="Overview")
        # # viewT1.style.button_color = 'lightblue'
        # # viewT1.on_click(on_viewT1_clicked)
        # hboxTeMHL = widgets.HBox([teMHL, selectTeMHL],layout=widgets.Layout( border='dotted black 1px',justify_content = 'space-between'))
        #                                                        # height='40px', align_items='center',   justify_content = 'space-between',
        #                                                        # margin='0px 0px 0px 10px'))
        # TEMPLATES.append(hboxTeMHL)
        #

        vb = widgets.VBox([headerT]+TEMPLATES)#+TUTORIALS)
        return vb

    def home_page_widgets(self):
        def on_selectN_clicked(b):
            self.block_id = 1   #restart the block ids
            self.output_directory = outdir.value
            rm = [i for i in self.pages if i not in [0,1]]
            for j in rm:
                del self.pages[j]
            self.comp_graph = []

            ## clear ouput and update the graph viz
            self.graph.close()
            # dot = Digraph(format='png')
            # for edge in self.comp_graph:
            #     dot.node('%i' % edge[0], label='%i %s' % (edge[0], self.pages[edge[0]].title))
            #     dot.node('%i' % edge[2], label='%i %s' % (edge[2],self.pages[edge[2]].title))
            #     dot.edge('%i' % edge[0], '%i'%edge[2], label='%s > %s' % (edge[1], edge[3]), labelfontcolor='green')
            # self.graph = widgets.Image(value=dot.pipe(),format='png')
            # display(self.graph)

            self.add_page()

        def on_selectE_clicked(b):
            if txtarea.value is not '':
                old = [i for i in self.pages]
                try:
                    self.parser(txtarea.value.split('\n'))
                    # update the current_bid
                    self.block_id = max(self.pages)
                    selectE.icon = 'check'
                except Exception as err:
                    print "Invalid configuration file ..."
                    print "    IOError: %s"%err.message
                    print "... Not loaded!"
                    selectE.icon = 'remove'
                    rm = [i for i in self.pages if i not in old]
                    for ib in rm:
                        if ib in self.pages:
                            del self.pages[ib]

                self.debut = False
                self.add_page()

                ## clear ouput and update the graph viz
                self.graph.close()
                dot = Digraph(format='png')
                for edge in self.comp_graph:
                    dot.node('%i' % edge[0], label='%i %s' % (edge[0], self.pages[edge[0]].title))
                    dot.node('%i' % edge[2], label='%i %s' % (edge[2], self.pages[edge[2]].title))
                    dot.edge('%i' % edge[0], '%i' % edge[2], label='%s > %s' % (edge[1], edge[3]), labelfontcolor='green')
                self.graph = widgets.Image(value=dot.pipe(), format='png')
                display(self.graph)

        def on_show_clicked(b):
            lines = self.configurer()
            txt.value = ''.join(lines)

        def on_save_clicked(b):
            path = self.check_filename(filename.value)
            lines = self.configurer()
            try:
                if len(lines) >0:
                    with open(path, 'w') as config:
                        for line in lines:
                            config.write(line)
                    print "\nThe ChemML Wrapper's config file has been successfully saved ..."
                    print "    config file path: %s" % path
                    print "    current directory: %s" % os.getcwd()
                    print "    what's next? run the ChemML Wrapper using the config file with the following codes:"
                    print "        >>> from cheml import wrapperRUN"
                    print "        >>> wrapperRUN(INPUT_FILE = 'path_to_the_config_file', OUTPUT_DIRECTORY = '%s')" % outdir.value
                    print "... you can also create a python script of the above codes and run it on any cluster that ChemML is installed."
                    save.icon = 'check'
                else:
                    print "The config file is empty ..."
                    print "... Not saved!"
                    save.icon = 'remove'
            except:
                print "The config file path or name is not supported ..."
                print "... Not saved!"
                save.icon = 'remove'

        header = widgets.Label(value='Choose how to start:', layout=widgets.Layout(width='50%'))
        # Tab: new script
        style = {'description_width': 'initial'}
        outdir = widgets.Text(
            value='CMLWrapper.out',
            placeholder='Type something',
            description='Output directory:',
            disabled=False,
            style = style,
            layout = widgets.Layout(margin='30px 0px 10px 0px'))
        headerN = widgets.Label(value='Start with a new script', layout=widgets.Layout(width='50%'))
        selectN = widgets.Button(description="Start", layout=widgets.Layout(margin='20px 0px 10px 115px'))
        selectN.style.button_color = 'lightblue'
        selectN.on_click(on_selectN_clicked)

        line = widgets.HBox([],layout=widgets.Layout(height='0px', border='dotted black 1px',margin='20px 0px 0px 0px'))
        note = widgets.HTML(value="<b>Note:</b> Don't forget to print/save the ChemML Wrapper's input file (configuration script) when you are done with editing:",
                            layout=widgets.Layout(margin='10px 0px 0px 0px'))
        show = widgets.Button(description="Print Script", layout=widgets.Layout(width= "120px",margin='10px 0px 0px 0px'))
        show.style.button_color = 'lightblue'
        show.on_click(on_show_clicked)
        txt = widgets.Textarea(
            placeholder="Press the 'Print Script' button to print the ChemML Wrapper's configuration script here. Copy and save it for the future use.",
            disabled=False,
            layout=widgets.Layout(width='100%',margin='0px 0px 0px 10px'))
        hbox1 = widgets.HBox([show,txt],layout=widgets.Layout(margin='10px 0px 0px 0px',align_items='center', justify_content = 'center'))

        save = widgets.Button(description="Save Script", layout=widgets.Layout(width= "120px",margin='20px 0px 0px 0px'))
        save.style.button_color = 'lightblue'
        save.on_click(on_save_clicked)
        # f_label = widgets.Label('config file name:',layout=widgets.Layout(width='100%',margin='20px 0px 0px 20px'))
        style = {'description_width': 'initial'}
        filename = widgets.Text(
            value='cheml_config.txt',
            description = 'config file name:',
            placeholder='cheml_config.txt',
            disabled=False,
            style= style,
            layout=widgets.Layout(margin='20px 0px 0px 20px'))
        hbox2 = widgets.HBox([save, filename],layout=widgets.Layout(margin='10px 0px 0px 0px',align_items='center', justify_content = 'center'))

        vboxN = widgets.VBox([headerN,outdir,selectN])
        # Tab: existing script
        headerE = widgets.Label(value='Load an existing script', layout=widgets.Layout(width='50%'))
        txtarea = widgets.Textarea(
            placeholder='copy a ChemML script (config file) here',
            disabled=False,
            layout = widgets.Layout(width='50%'))
        selectE = widgets.Button(description="Load", layout=widgets.Layout(margin='20px 0px 10px 115px'))
        selectE.style.button_color = 'lightblue'
        selectE.on_click(on_selectE_clicked)
        vboxE = widgets.VBox([headerE, txtarea, selectE])
        # Tab: template
        vboxT = self.templates_widgets()

        tabs = widgets.Tab()
        tabs.children = [vboxN, vboxE, vboxT]
        tabs.set_title(0, 'New script')
        tabs.set_title(1, 'Existing script')
        tabs.set_title(2, 'Template workflow')

        self.home_page_VBox = widgets.VBox([header, tabs,line,note,hbox1, hbox2])

    def home_page(self):
        id = 0
        # clear_output()
        self.home_page_widgets()
        self.pages[id] = container_page('Home page', id, self.home_page_VBox)
        self.display_accordion(id)
        display(self.accordion)
        print 'The computation graph will be displayed here:'
        display(self.graph)

    ################################

    def add_page_widgets(self):
        def on_select_clicked(b):
            self.block_id += 1
            block_params = {'task': task_w.value, 'subtask': subtask_w.value, \
                                          'host': host_w.value, 'function': func_w.value, \
                                          'wparams':{}, 'fparams':{}, 'inputs':{}, 'outputs':{}}
            self.pages[self.block_id] = container_page('%s'%func_w.value, self.block_id, None, block_params)
            self.current_bid = self.block_id
            self.debut = True
            self.custom_function_page()

        def _subtask_update():
            subtask_opts = [i for i in self.combinations[task_w.value]]
            subtask_w.options = subtask_opts
            subtask_w.value = subtask_opts[0]

        def _host_update():
            host_opts = [i for i in self.combinations[task_w.value][subtask_w.value]]
            host_w.options = host_opts
            host_w.value = host_opts[0]

        def _func_update():
            func_opts = [i for i in self.combinations[task_w.value][subtask_w.value][host_w.value]]
            func_w.options = func_opts
            func_w.value = func_opts[0]

        def handle_task_change(t):
            _subtask_update()
            _host_update()
            _func_update()

        def handle_subtask_change(s):
            _host_update()
            _func_update()

        def handle_host_change(h):
            _func_update()

        header = widgets.Label(value='Choose a method:', layout=widgets.Layout(width='50%'))
        task_options = self.tasks[0:5] + self.tasks[6:8]
        task_w = widgets.Dropdown(
            options=task_options,
            value=task_options[0],
            description='Task:')
        subtask_options = [i for i in self.combinations[task_w.value]]
        subtask_w = widgets.Dropdown(
            options=subtask_options,
            value=subtask_options[0],
            description='Subtask:')
        host_options = [i for i in self.combinations[task_w.value][subtask_w.value]]
        host_w = widgets.Dropdown(
            options=host_options,
            value=host_options[0],
            description='Host:')
        func_options = [i for i in self.combinations[task_w.value][subtask_w.value][host_w.value]]
        func_w = widgets.Dropdown(
            options=func_options,
            value=func_options[0],
            description='Function:')
        select = widgets.Button(description="Select", layout=widgets.Layout(margin='20px 0px 10px 115px'))
        select.style.button_color = 'lightblue'
        select.on_click(on_select_clicked)

        task_w.observe(handle_task_change, names='value')
        subtask_w.observe(handle_subtask_change, names='value')
        host_w.observe(handle_host_change, names='value')

        add_page_VBox = widgets.VBox([header, task_w, subtask_w, host_w, func_w, select])

        return add_page_VBox

    def add_page(self):
        id = 1
        add_page_VBox = self.add_page_widgets()
        self.pages[id] = container_page('Add a block', id, add_page_VBox)
        self.display_accordion(id)

    ################################

    def db_extract_function(self, host, function):
        if host == 'sklearn':
            metadata = getattr(sklearn_db, function)()
        elif host == 'cheml':
            metadata = getattr(cheml_db, function)()
        elif host == 'pandas':
            metadata = getattr(pandas_db, function)()
        wparams = {i:copy.deepcopy(vars(metadata.WParameters)[i]) for i in vars(metadata.WParameters).keys() if
                     i not in ('__module__', '__doc__')}
        fparams = {i:copy.deepcopy(vars(metadata.FParameters)[i]) for i in vars(metadata.FParameters).keys() if
                     i not in ('__module__', '__doc__')}
        inputs = {i: copy.deepcopy(vars(metadata.Inputs)[i]) for i in vars(metadata.Inputs).keys() if
                  i not in ('__module__', '__doc__')}
        outputs = {i: copy.deepcopy(vars(metadata.Outputs)[i]) for i in vars(metadata.Outputs).keys() if
                   i not in ('__module__', '__doc__')}
        return wparams, fparams, inputs, outputs, metadata

    def custom_function_IO_w(self):
        if not self.debut:
            self.current_bid = sorted(self.pages)[self.accordion.selected_index]

        def remove_nodes_toSEND(n = self.current_bid, remove_bids = set()):
            # find all the nodes that the current node can not send info to them
            remove_bids.add(n)  # including itself
            for e in self.comp_graph:
                if e[2]==n:
                    remove_bids.add(e[0])
                    remove_nodes_toSEND(n=e[0], remove_bids=remove_bids)
            return remove_bids

        def remove_nodes_toRECV(n = self.current_bid, remove_bids = set()):
            # find all the nodes that the current node can not receive info from them
            remove_bids.add(n)  # including itself
            for e in self.comp_graph:
                if e[0]==n:
                    remove_bids.add(e[2])
                    remove_nodes_toRECV(n=e[2], remove_bids=remove_bids)
            return remove_bids

        def refresh_tasks():
            if not self.debut:
                self.current_bid = sorted(self.pages)[self.accordion.selected_index]

            ### Senders
            ## block ids that can receive info from the current block
            bidS_options = [i for i in self.pages if i not in [0, 1]]
            rm_n_2S = remove_nodes_toSEND(self.current_bid,set())
            for n in rm_n_2S:
                bidS_options.remove(n)
            bidS.options = sorted(bidS_options)
            if len(bidS_options)>0:
                bidS.value = sorted(bidS_options)[0]
                external_inputs = [token for token in self.pages[bidS.value].block_params['inputs']]
                # remove input tokens that are already taken
                rm = []
                for t in external_inputs:
                    for e in self.comp_graph:
                        if e[2:] == (bidS.value, t):
                            rm.append(t)
                for token in rm:
                    external_inputs.remove(token)
                external_receivers.options = sorted(external_inputs)
                if len(external_inputs)>0:
                    external_receivers.value = sorted(external_inputs)[0]
            else:
                external_receivers.options = []

            ### Receivers
            ## update input tokens to receive only once
            input_tokens = [token for token in self.pages[self.current_bid].block_params['inputs']]
            # remove input tokens that are already taken
            rm = []
            for t in input_tokens:
                for e in self.comp_graph:
                    if e[2:] == (self.current_bid, t):
                        rm.append(t)
            for token in rm:
                input_tokens.remove(token)
            receiver.options = sorted(input_tokens)
            if len(input_tokens)>0:
                receiver.value = sorted(input_tokens)[0]
                # block ids that can send info to the current block
                bidR_options = [i for i in self.pages if i not in [0, 1]]
                rm = remove_nodes_toRECV(self.current_bid,set())
                for n in rm:
                    bidR_options.remove(n)
                bidR.options = sorted(bidR_options)
                if len(bidR_options)>0:
                    bidR.value = sorted(bidR_options)[0]
                    external_outputs = [token for token in self.pages[bidR.value].block_params['outputs']]
                    external_senders.options = external_outputs
                else:
                    external_senders.options = []
            else:
                bidR.options = []
                external_senders.options = []

            ## update pipes options from comp_graph
            ps = ["%i, %s      >>>      %i, %s" % (e[0], e[1], e[2], e[3]) for e in self.comp_graph if self.current_bid in e]
            pipes.options = sorted(ps)

        def display_graph():
            ## clear ouput and update the graph viz
            self.graph.close()
            dot = Digraph(format='png')
            for edge in self.comp_graph:
                dot.node('%i' % edge[0], label='%i %s' % (edge[0], self.pages[edge[0]].title))
                dot.node('%i' % edge[2], label='%i %s' % (edge[2],self.pages[edge[2]].title))
                dot.edge('%i' % edge[0], '%i'%edge[2], label='%s > %s' % (edge[1], edge[3]), labelfontcolor='green')
            self.graph = widgets.Image(value=dot.pipe(),format='png')
            display(self.graph)

        def on_refresh_clicked(b):
            if not self.debut:
                self.current_bid = sorted(self.pages)[self.accordion.selected_index]
            refresh_tasks()

        def on_addS_clicked(b):
            if not self.debut:
                self.current_bid = sorted(self.pages)[self.accordion.selected_index]
            if sender.value is not None and bidS.value is not None and external_receivers.value is not None:
                input_type = set(self.pages[bidS.value].block_params['inputs'][external_receivers.value].types)
                output_type = set(self.pages[self.current_bid].block_params['outputs'][sender.value].types)
                if input_type.issubset(output_type) or output_type.issubset(input_type):
                    all_receivers = [e[2:] for e in self.comp_graph]
                    if (bidS.value, external_receivers.value) not in all_receivers:
                        edge = (self.current_bid, sender.value, bidS.value, external_receivers.value)
                        if edge not in self.comp_graph:
                            self.comp_graph.append(edge)
            refresh_tasks()
            display_graph()

        def on_addR_clicked(b):
            if not self.debut:
                self.current_bid = sorted(self.pages)[self.accordion.selected_index]
            if receiver.value is not None and bidR.value is not None and external_senders.value is not None:
                output_type = set(self.pages[bidR.value].block_params['outputs'][external_senders.value].types)
                input_type = set(self.pages[self.current_bid].block_params['inputs'][receiver.value].types)
                if input_type.issubset(output_type) or output_type.issubset(input_type):
                    all_receivers = [e[2:] for e in self.comp_graph]
                    if (self.current_bid, receiver.value) not in all_receivers:
                        edge = (bidR.value, external_senders.value, self.current_bid, receiver.value)
                        if edge not in self.comp_graph:
                            self.comp_graph.append(edge)
            ## all refresh tasks:
            refresh_tasks()
            display_graph()

        def bidS_value_change(change):
            if not self.debut:
                self.current_bid = sorted(self.pages)[self.accordion.selected_index]
            if bidS.value is not None:
                external_inputs = [token for token in self.pages[bidS.value].block_params['inputs']]
                rm = []
                for t in external_inputs:
                    for e in self.comp_graph:
                        if e[2:] == (bidS.value, t):
                            rm.append(t)
                for token in rm:
                    external_inputs.remove(token)
                external_receivers.options = sorted(external_inputs)
                if len(external_inputs)>0:
                    external_receivers.value = sorted(external_inputs)[0]
            else:
                external_receivers.options = []

        def bidR_value_change(change):
            if not self.debut:
                self.current_bid = sorted(self.pages)[self.accordion.selected_index]
            if bidR.value is not None:
                external_outputs = [token for token in self.pages[bidR.value].block_params['outputs']]
                external_senders.options = sorted(external_outputs)
                if len(external_outputs) > 0:
                    external_senders.value = sorted(external_outputs)[0]
            else:
                external_senders.options = []

        def on_remove_clicked(b):
            if not self.debut:
                self.current_bid = sorted(self.pages)[self.accordion.selected_index]
            rm = pipes.value
            for p in rm:
                p = p.strip().split(',')
                m = p[1].strip().split('      >>>      ')
                edge = (int(p[0]), m[0].strip(), int(m[1]), p[2].strip())
                if edge in self.comp_graph:
                    self.comp_graph.remove(edge)
            refresh_tasks()
            display_graph()


        self.pages[self.current_bid].IO_refresher = refresh_tasks

        ## Sender
        output_tokens = [token for token in self.pages[self.current_bid].block_params['outputs']]
        headerS = widgets.HTML(value='<b> Send >>> </b>', layout=widgets.Layout(width='50%',margin='10px 0px 0px 10px'))
        # listS = widgets.HTML(value='output tokens: %s'%', '.join(sorted(output_tokens)), \
        #                      layout=widgets.Layout(margin='3px 0px 0px 20px'))
        sender = widgets.Dropdown(
            options=sorted(output_tokens),
            # value=output_tokens[0],
            description='output token:')
        hbox1S = widgets.HBox([sender],layout=widgets.Layout(height='40px', border='dotted black 1px',
                                                               align_items='center',  # justify_content = 'center',
                                                               margin='0px 0px 0px 10px'))
        toS = widgets.HTML(value='<b> >>> </b>')
        bidS_options = [i for i in self.pages if i not in [0, 1]]
        rm_n_2S = remove_nodes_toSEND(self.current_bid,set())
        for n in rm_n_2S:
            bidS_options.remove(n)
        bidS = widgets.Dropdown(
            options=sorted(bidS_options),
            description='block#:',
            layout=widgets.Layout(width='140px'))
        bidS.observe(bidS_value_change,names='value')
        refreshS = widgets.Button(icon='refresh', layout=widgets.Layout(width='40px'))
        # refreshS.style.button_color = 'darkseagreen'
        refreshS.on_click(on_refresh_clicked)
        if bidS.value is not None:
            external_inputs = [token for token in self.pages[bidS.value].block_params['inputs']]
            rm = []
            for t in external_inputs:
                for e in self.comp_graph:
                    if e[2:] == (bidS.value, t):
                        rm.append(t)
            for token in rm:
                external_inputs.remove(token)
        else:
            external_inputs = []
        external_receivers = widgets.Dropdown(
            options = sorted(external_inputs),
            description='input token:')
        hbox2S = widgets.HBox([bidS,refreshS, external_receivers],
                             layout=widgets.Layout(height='40px', border='dotted black 1px',
                                                   align_items='center', justify_content='center'))
        addS = widgets.Button(description="Add", layout=widgets.Layout(width='60px', margin='0px 10px 0px 0px'))
        addS.style.button_color = 'lightblue'
        addS.on_click(on_addS_clicked)
        hboxS = widgets.HBox([hbox1S, toS, hbox2S, addS],
                             layout=widgets.Layout(justify_content='space-between',
                                                   align_items='center',
                                                   margin='10px 0px 20px 0px'))

        ## Receiver
        note = widgets.HTML(value='<b> Note: </b> This page automatically avoid: (1) loops, (2) type inconsistency, and (3) more than one input per token.', layout=widgets.Layout(margin='20px 0px 0px 10px'))
        headerR = widgets.HTML(value='<b> Receive <<< </b>', layout=widgets.Layout(width='50%',margin='10px 0px 0px 10px'))
        input_tokens = [token for token in self.pages[self.current_bid].block_params['inputs']]
        listR = widgets.HTML(value='all input tokens: %s'%', '.join(sorted(input_tokens)),
                             layout=widgets.Layout(margin='0px 0px 0px 20px'))
        rm = []
        for t in input_tokens:
            for e in self.comp_graph:
                if e[2:] == (self.current_bid, t):
                    rm.append(t)
        for token in rm:
            input_tokens.remove(token)
        receiver = widgets.Dropdown(
            options=input_tokens,
            # value=input_tokens[0],
            disabled = False,
            description = 'input token:')
        hbox1R = widgets.HBox([receiver],layout=widgets.Layout(height='40px', border='dotted black 1px',
                                                               align_items='center',  # justify_content = 'center',
                                                               margin='0px 0px 0px 10px'))
        fro = widgets.HTML(value='<b> <<< </b>')
        bidR_options = [i for i in self.pages if i not in [0, 1]]
        rm = remove_nodes_toRECV(self.current_bid, set())
        for n in rm:
            bidR_options.remove(n)
        bidR = widgets.Dropdown(
            options = sorted(bidR_options),
            description='block#:',
            disabled=False,
            layout=widgets.Layout(width='140px'))
        bidR.observe(bidR_value_change,names='value')
        refresh = widgets.Button(icon='refresh',disabled=False, layout=widgets.Layout(width='40px'))
        # refresh.style.button_color = 'darkseagreen'
        refresh.on_click(on_refresh_clicked)
        if bidR.value is not None:
            external_outputs = [token for token in self.pages[bidR.value].block_params['outputs']]
        else:
            external_outputs = []
        external_senders = widgets.Dropdown(
            options = sorted(external_outputs),
            disabled=False,
            description='output token:')
        hbox2R = widgets.HBox([bidR, refresh,external_senders],
                             layout=widgets.Layout(height='40px', border='dotted black 1px',
                                                   align_items='center', justify_content='center'))
        addR = widgets.Button(description="Add", disabled=False, layout=widgets.Layout(width='60px', margin='0px 10px 0px 0px'))
        addR.style.button_color = 'lightblue'
        addR.on_click(on_addR_clicked)
        hboxR = widgets.HBox([hbox1R, fro, hbox2R, addR],
                             layout=widgets.Layout(justify_content='space-between',
                                                   align_items='center',
                                                   margin='10px 0px 20px 0px'))

        ## vbox
        pipes_options = [edge for edge in self.comp_graph if self.current_bid in edge]
        pipes = widgets.SelectMultiple(
            options= sorted(pipes_options),
            # description='list of pipes:',
            layout=widgets.Layout(margin='20px 0px 20px 250px'))
        remove = widgets.Button(description="Remove pipe", icon = 'remove',
                                layout=widgets.Layout(width='120px', margin='40px 0px 0px 10px'))
        remove.style.button_color = 'lightblue'
        remove.on_click(on_remove_clicked)
        hbox4 = widgets.HBox([pipes, remove], margin='0px 0px 0px 100px')

        IO_vbox = widgets.VBox([note, headerS, hboxS, headerR, listR, hboxR, hbox4])#, layout=widgets.Layout(border='solid darkslategray 1px'))
        return IO_vbox

    def custom_function_params_w(self):
        # def handle_param_change(b):
        #     for item in self.pages[current_bid].block_params['wparams']:
        #         self.pages[current_bid].block_params['wparams'][item].value = self.pages[current_bid].block_params['wparams'][item].widget.value
        #     for item in self.pages[current_bid].block_params['fparams']:
        #         self.pages[current_bid].block_params['fparams'][item].value = self.pages[current_bid].block_params['fparams'][item].widget.value
        if not self.debut:
            self.current_bid = sorted(self.pages)[self.accordion.selected_index]

        def on_default_clicked(b):
            if not self.debut:
                self.current_bid = sorted(self.pages)[self.accordion.selected_index]
            for item in self.pages[self.current_bid].block_params['wparams']:
                self.pages[self.current_bid].block_params['wparams'][item].widget.value = \
                str(self.pages[self.current_bid].block_params['wparams'][item].default)
            for item in self.pages[self.current_bid].block_params['fparams']:
                self.pages[self.current_bid].block_params['fparams'][item].widget.value = \
                str(self.pages[self.current_bid].block_params['fparams'][item].default)

        # wrapper parameters widgets
        header = widgets.HTML(value='<b>Wrapper parameters:</b>', layout=widgets.Layout(width='50%'))
        wparams = self.pages[self.current_bid].block_params['wparams']
        wparams_boxes = []
        for item in sorted(wparams):
            style = {'description_width': 'initial'}
            wp = widgets.Text(
                value=str(wparams[item].default),
                placeholder=str(wparams[item].options),
                description=wparams[item].name,
                disabled=False,
                style = style,
                )
            wparams[item].widget = wp
            # wp.observe(handle_param_change, names='value')
            wp_checker = widgets.Checkbox(value=False,
                                          description='Check to set',
                                          disabled=False,
                                          indent=False,
                                          layout = widgets.Layout(margin='0px 0px 0px 10px')
            )
            if self.pages[self.current_bid].block_params['wparams'][item].required:
                wp_checker.value = True
            wparams[item].checker = wp_checker

            wformat = widgets.Text(
                value=str(wparams[item].format),
                description= 'Type:',
                disabled=True)

            hwp = widgets.Box([wp])
            if self.pages[self.current_bid].block_params['wparams'][item].required:
                hwp.layout = widgets.Layout(border='dotted red 1px')

            hbox = widgets.HBox([hwp, wp_checker, wformat],)
                                # layout=widgets.Layout(display='flex',
                                #                       flex_flow='row',
                                #                       align_items='stretch'))
            wparams_boxes.append(hbox)
        wparams_vbox = widgets.VBox([header]+ wparams_boxes)
        self.pages[self.current_bid].block_params['wparams'] = wparams

        # function parameters widgets
        header = widgets.HTML(value='<b>Function parameters:</b>',
                              layout=widgets.Layout(width='50%',margin='20px 0px 0px 0px'))
        fparams = self.pages[self.current_bid].block_params['fparams']
        fparams_boxes = []
        for item in sorted(fparams):
            style = {'description_width': 'initial'}
            wp = widgets.Text(
                value=str(fparams[item].default),
                placeholder=str(fparams[item].options),
                description=fparams[item].name,
                disabled=False,
                style = style,
                )
            fparams[item].widget = wp
            # wp.observe(handle_param_change, names='value')
            wp_checker = widgets.Checkbox(value=False,
                                          description='Check to set',
                                          disabled=False,
                                          indent = False,
                                          layout = widgets.Layout(margin='0px 0px 0px 10px')
                                         )

            if self.pages[self.current_bid].block_params['fparams'][item].required:
                wp_checker.value = True
            fparams[item].checker = wp_checker

            wformat = widgets.Text(
                value=str(fparams[item].format),
                description='Type:',
                disabled=True,
                layout = widgets.Layout(width='40%'))

            hwp = widgets.Box([wp])
            if self.pages[self.current_bid].block_params['fparams'][item].required:
                hwp.layout = widgets.Layout(border='dotted red 1px')


            hbox = widgets.HBox([ hwp, wp_checker, wformat],)
                                # layout=widgets.Layout(display='flex',
                                #                       flex_flow='row',))
                                                      # align_items='stretch'))
            fparams_boxes.append(hbox)
        fparams_vbox = widgets.VBox([header]+ fparams_boxes)
        self.pages[self.current_bid].block_params['fparams'] = fparams

        defaultB = widgets.Button(description="Default values",
                                  layout=widgets.Layout(width='110px',margin='10px 0px 10px 180px'))
        defaultB.style.button_color = 'lightblue'
        defaultB.on_click(on_default_clicked)

        params_vbox = widgets.VBox([wparams_vbox,fparams_vbox,defaultB],
                                   layout=widgets.Layout(justify_content='center', align_content='center'))

        return params_vbox

    def set_default_params_IO(self):
        if not self.debut:
            self.current_bid = sorted(self.pages)[self.accordion.selected_index]
        host = self.pages[self.current_bid].block_params['host']
        function = self.pages[self.current_bid].block_params['function']
        wparams, fparams, inputs, outputs, metadata = self.db_extract_function(host,function)
        self.pages[self.current_bid].block_params['wparams'] = wparams
        self.pages[self.current_bid].block_params['fparams'] = fparams
        self.pages[self.current_bid].block_params['inputs'] = inputs
        self.pages[self.current_bid].block_params['outputs'] = outputs
        self.pages[self.current_bid].block_params['requirements'] = metadata.requirements
        self.pages[self.current_bid].block_params['documentation'] = metadata.documentation
        self.pages[self.current_bid].block_params['task'] = metadata.task
        self.pages[self.current_bid].block_params['subtask'] = metadata.subtask

    def custom_function_widgets(self):
        if not self.debut:
            self.current_bid = sorted(self.pages)[self.accordion.selected_index]

        def on_remove_clicked(b):
            if not self.debut:
                self.current_bid = sorted(self.pages)[self.accordion.selected_index]
            del self.pages[self.current_bid]
            rm = [edge for edge in self.comp_graph if self.current_bid in edge]
            for e in rm:
                self.comp_graph.remove(e)

            ## clear ouput and update the graph viz
            self.graph.close()
            dot = Digraph(format='png')
            for edge in self.comp_graph:
                dot.node('%i' % edge[0], label='%i %s' % (edge[0], self.pages[edge[0]].title))
                dot.node('%i' % edge[2], label='%i %s' % (edge[2],self.pages[edge[2]].title))
                dot.edge('%i' % edge[0], '%i'%edge[2], label='%s > %s' % (edge[1], edge[3]), labelfontcolor='green')
            self.graph = widgets.Image(value=dot.pipe(),format='png')
            display(self.graph)


            self.add_page()

        # initialize, only if fparams, inputs and ouputs are empty
        if len(self.pages[self.current_bid].block_params['fparams']) == len(self.pages[self.current_bid].block_params['inputs']) ==\
                len(self.pages[self.current_bid].block_params['outputs']) == 0:
            self.set_default_params_IO()
        # info
        t = self.pages[self.current_bid].block_params['task']
        s = self.pages[self.current_bid].block_params['subtask']
        h = self.pages[self.current_bid].block_params['host']
        f = self.pages[self.current_bid].block_params['function']
        tshfTB = widgets.ToggleButtons(
            options=[t, s, h, f],
            disabled=False,
            button_style='info',  # 'success', 'info', 'warning', 'danger' or ''
            layout=widgets.Layout(margin='0px 0px 0px 180px'))
        reqL = widgets.HTML(value="<b>Requirements:  </b>"+str(self.pages[self.current_bid].block_params['requirements']),
                            layout=widgets.Layout(width='100%'))
        docL = widgets.HTML(value="<b>Documentation: </b>"+self.pages[self.current_bid].block_params['documentation'],
                            layout=widgets.Layout(width='100%'))

        # parameters widgets
        params_vbox = self.custom_function_params_w()

        # inputs and outputs
        IO_vbox = self.custom_function_IO_w()

        removeB = widgets.Button(description="Remove block",layout=widgets.Layout(margin='10px 0px 10px 400px'))
        removeB.style.button_color = 'lightblue'
        removeB.on_click(on_remove_clicked)

        custom_f_accordion = widgets.Tab(layout=widgets.Layout(margin='10px 0px 10px 0px'))
        custom_f_accordion.children = [params_vbox, IO_vbox]
        custom_f_accordion.set_title(0, 'Parameters')
        custom_f_accordion.set_title(1, 'Input/Output')
        # custom_f_accordion.selected_index = 0

        custom_f_vbox = widgets.VBox([tshfTB, reqL, docL, custom_f_accordion, removeB],
                                     layout=widgets.Layout(justify_content='center', align_content='center'))

        return custom_f_vbox

    def custom_function_page(self):
        custom_function_VBox = self.custom_function_widgets()
        self.pages[self.current_bid].widget = custom_function_VBox
        # display
        active_id = [i for i in sorted(self.pages)].index(self.current_bid)
        active_id = 1 # keep it on the add page (if comment out this line the newly added block will become active)
        self.display_accordion(active_id)
        self.debut = False

    ################################

    def parser(self, script):
        """
        The main funtion for parsing cheml script.
        It starts with finding blocks and then runs other functions.

        :return:
        cmls: pages
        """
        blocks={}
        it = max(self.pages)+1  # must be +2 but there is another +1 in the loop

        check_block = False
        for line in script:
            if '##' in line:
                it += 1
                blocks[it] = [line]
                check_block = True
                continue
            elif '#' in line:
                check_block = False
            elif check_block and ('<' in line or '>' in line):
                blocks[it].append(line)
                continue
        self.loading_bids = range(min(blocks),max(blocks)+1)
        self._options(blocks)
        # to make self.comp_graph with cmls (onlly requires send and receive)
        self._transform()

    def _db_extract_function(self, host, function):
        if host == 'sklearn':
            metadata = getattr(sklearn_db, function)()
        elif host == 'cheml':
            metadata = getattr(cheml_db, function)()
        wparams = {i:copy.deepcopy(vars(metadata.WParameters)[i]) for i in vars(metadata.WParameters).keys() if
                     i not in ('__module__', '__doc__')}
        fparams = {i:copy.deepcopy(vars(metadata.FParameters)[i]) for i in vars(metadata.FParameters).keys() if
                     i not in ('__module__', '__doc__')}
        inputs = {i: copy.deepcopy(vars(metadata.Inputs)[i]) for i in vars(metadata.Inputs).keys() if
                  i not in ('__module__', '__doc__')}
        outputs = {i: copy.deepcopy(vars(metadata.Outputs)[i]) for i in vars(metadata.Outputs).keys() if
                   i not in ('__module__', '__doc__')}
        return wparams, fparams, inputs, outputs, metadata

    def _functions(self, line):
        if '<' in line:
            function = line[line.index('##')+2:line.index('<')].strip()
        elif '>' in line:
            function = line[line.index('##')+2:line.index('>')].strip()
        else:
            function = line[line.index('##')+2:].strip()
        return function

    def _parameters(self, block,item):
        parameters = {}
        send = {}
        recv = {}
        for line in block:
            while '<<' in line:
                line = line[line.index('<<')+2:].strip()
                if '<' in line:
                    args = line[:line.index('<')].strip()
                else:
                    args = line.strip()
                param = args[:args.index('=')].strip()
                val = args[args.index('=')+1:].strip()
                parameters[param] = val #value(val) #"%s"%val
            while '>>' in line:
                line = line[line.index('>>') + 2:].strip()
                if '>' in line:
                    args = line[:line.index('>')].strip()
                else:
                    args = line.strip()
                arg = args.split()
                if len(arg) == 2:
                    a, b = arg
                    if isint(b) and not isint(a):
                        send[(a, int(b))] = item
                    elif isint(a) and not isint(b):
                        recv[(b, int(a))] = item
                    else:
                        msg = 'wrong format of send and receive in block #%i at %s (send: >> var id; recv: >> id var)' % (item+1, args)
                        raise IOError(msg)
                else:
                    msg = 'wrong format of send and receive in block #%i at %s (send: >> var id; recv: >> id var)'%(item+1,args)
                    raise IOError(msg)
        return parameters, send, recv

    def _options(self, blocks):
        for bid in self.loading_bids:
            block = blocks[bid]
            # task = self._functions(block[0])
            parameters, send, recv = self._parameters(block,bid)
            host = parameters['host']
            function = parameters['function']
            # here we also add send and receive to be handled in the self.transform for making the comp_graph
            block_params = {'task': '', 'subtask': '', \
                            'host': host, 'function': function, \
                            'wparams': {}, 'fparams': {}, 'inputs': {}, 'outputs': {}, 'send':send, 'recv':recv}
            self.pages[bid] = container_page('%s'%function, bid, None, block_params)
            self.current_bid = bid
            self.debut = True
            # to make all the widgets (store widget in the page.widget and parameters and IOs db containers
            custom_function_VBox = self.custom_function_widgets()
            self.pages[bid].widget = custom_function_VBox

            # change the widget values to the values from config file
            for param in parameters:
                if param not in ['host', 'function']:
                    if param in self.pages[bid].block_params['wparams']:
                        self.pages[bid].block_params['wparams'][param].widget.value = parameters[param]
                        # checkbox if value is different from default
                        if self.pages[bid].block_params['wparams'][param].default != parameters[param]:
                            self.pages[bid].block_params['wparams'][param].checker.value = True
                    elif param in self.pages[bid].block_params['fparams']:
                        self.pages[bid].block_params['fparams'][param].widget.value = parameters[param]
                        # checkbox if value is different from default
                        if self.pages[bid].block_params['fparams'][param].default != parameters[param]:
                            self.pages[bid].block_params['fparams'][param].checker.value = True
                    else:
                        msg = "(host: %s, function: %s): Not a valid parameter '%s'"%(host, function, param)
                        raise IOError(msg)

    def _transform(self):
        """
        goals:
            - collect all sends and receives
            - check send and receive format.
            - make the computational graph

        """
        send_all = []
        recv_all = []
        for bid in self.loading_bids:
            send_all += self.pages[bid].block_params['send'].items()
            recv_all += self.pages[bid].block_params['recv'].items()
        # check send and recv
        if len(send_all) > len(recv_all):
            msg = '@cheml script - number of sent tokens must be less or equal to number of received tokens'
            raise ValueError(msg)
        send_ids = [k[1] for k,v in send_all]
        recv_ids = [k[1] for k,v in recv_all]
        for id in send_ids:
            if send_ids.count(id)>1:
                msg = 'identified non unique send id (id#%i)'%id
                raise NameError(msg)
        if set(send_ids) != set(recv_ids):
            print set(send_ids),set(recv_ids)
            msg = 'missing pairs of send and receive id:\n send IDs:%s\n recv IDs:%s\n'%(str(set(send_ids)),str(set(recv_ids)))
            raise ValueError(msg)

        # make graph
        reformat_send = {k[1]:[v,k[0]] for k,v in send_all}
        self.comp_graph += [tuple(reformat_send[k[1]]+[v,k[0]]) for k,v in recv_all]


# examples for dictionaries' fomrat:

# comp_graph = [[0, 'df', 2, 'df'], [0, 'api', 1, 'iv1'], [1, 'ov1', 3, 'df2'], [2, 'df', 3, 'df1'], [1, 'ov2', 4, 'df']]

# self.pages[ib].block_params = { 'function': 'mlp_hogwild',
#                                 'inputs': {'api': <cheml.wrappers.database.containers.Input object at 0x12351c510>,
#                                             'dfy': <cheml.wrappers.database.containers.Input object at 0x12421a610>,
#                                             'dfx': <cheml.wrappers.database.containers.Input object at 0x12421a690>},
#                                 'subtask': 'regression',
#                                 'task': 'Model',
#                                 'requirements': (('ChemML', '0.1.0'), ('scikit-learn', '0.19.0'), ('pandas', '0.20.3')),
#                                 'wparams': {'func_method': <cheml.wrappers.database.containers.Parameter object at 0x123bf1d50>},
#                                 'fparams': {'rms_decay': <cheml.wrappers.database.containers.Parameter object at 0x1235cc750>,
#                                             'learn_rate': <cheml.wrappers.database.containers.Parameter object at 0x1235cc250>},
#                                 'outputs': {'api': <cheml.wrappers.database.containers.Output object at 0x12421a6d0>,
#                                             'dfy_predict': <cheml.wrappers.database.containers.Output object at 0x12421a710>},
#                                 'documentation': '',
#                                 'host': 'cheml'
#                               }


# coding=utf-8
import sys
import types
import numpy as np
import pandas as pd
from ....data.materials.CompositionEntry import CompositionEntry
from ....data.materials.util.LookUpData import LookUpData
from ....utility.tools.OxidationStateGuesser import OxidationStateGuesser

class ChargeDependentAttributeGenerator:
    """Class to generate attributes derived from the oxidation states of
    elements in a material.

    Based on work by Deml et al.[1].

    Notes
    -----
    These features are based on the formal charges of materials determined
    using the OxidationStateGuesser. Currently implemented features:
    Statistics of formal charges (min, max, range, mean, variance)
    Cumulative ionization energies/ electron affinities
    Difference in electronegativities between cation and anion.

    For materials that the algorithm fails to find charge states, NaN is set
    for all features.

    References
    ----------
    .. [1] A. M. Deml, R. O’Hayre, C. Wolverton, and V. Stevanović,
    "Predicting density functional theory total energies and enthalpies of
    formation of metal-nonmetal compounds by linear regression," Physical
    Review B, vol. 93, no. 8, Feb. 2016.

    """

    def generate_features(self, entries):
        """
        Function to generate features as mentioned in the class description.

        Parameters
        ----------
        entries : array-like
            Compositions for which features are to be generated. A list of
            CompositionEntry's.

        Returns
        ----------
        features : DataFrame
            Features for the given entries. Pandas data frame containing the
            names and values of the descriptors.

        Raises
        ------
        ValueError
            If input is not of type list.
            If items in the list are not CompositionEntry instances.
        """

        # Initialize lists of feature values and headers for pandas data frame.
        feat_values = []
        feat_headers = []

        # Raise exception if input argument is not of type list of
        # CompositionEntry's.
        if (type(entries) is not types.ListType):
            raise ValueError("Argument should be of type list of "
                             "CompositionEntry's")
        elif (entries and not isinstance(entries[0], CompositionEntry)):
            raise ValueError("Argument should be of type list of "
                             "CompositionEntry's")

        # Insert feature headers here.
        n_features = 8
        feat_headers.append("min_Charge")
        feat_headers.append("max_Charge")
        feat_headers.append("maxdiff_Charge")
        feat_headers.append("mean_Charge")
        feat_headers.append("var_Charge")
        feat_headers.append("CumulativeIonizationEnergy")
        feat_headers.append("CumulativeElectronAffinity")
        feat_headers.append("AnionCationElectronegativityDiff")

        # Load properties here.
        en = LookUpData.load_property("Electronegativity")
        ea = LookUpData.load_property("ElectronAffinity")
        ie = LookUpData.load_property("IonizationEnergies")

        # Instantiate and initialize oxidation state guesser.
        ox_guesser = OxidationStateGuesser()
        ox_guesser.set_electronegativity(en)
        ox_guesser.set_oxidationstates(LookUpData.load_property(
            "OxidationStates"))

        missing_data = {}
        for entry in entries:
            tmp_list = []
            elems = entry.get_element_ids()
            fracs = entry.get_element_fractions()

            # Get possible states with charges.
            possible_states = ox_guesser.get_possible_states(entry)

            # If there are no possible states, set all features to NaN.
            if len(possible_states) == 0:
                for i in xrange(n_features):
                    tmp_list.append(np.nan)

                feat_values.append(tmp_list)
                continue

            # Check that we have data for all ionization energies.
            any_missing = False
            tmp_charges = possible_states[0]
            for i,elem in enumerate(elems):
                if len(ie[elem]) < tmp_charges[i]:
                    if elem not in missing_data:
                        missing_data[elem] = []
                    if possible_states[0][i] not in missing_data[elem]:
                        missing_data[elem].append(possible_states[0][i])
                    any_missing = True
                    break

            # Compute statistics related to charges.
            min_ = min(tmp_charges)
            max_ = max(tmp_charges)
            max_diff_ = max_ -  min_
            mean_ = np.average([abs(x) for x in tmp_charges], weights=fracs)
            var_ = np.average([abs(abs(x) - mean_) for x in tmp_charges],
                              weights=fracs)

            tmp_list.append(min_)
            tmp_list.append(max_)
            tmp_list.append(max_diff_)
            tmp_list.append(mean_)
            tmp_list.append(var_)

            if any_missing:
                tmp_list.append(np.nan)
                tmp_list.append(np.nan)
                tmp_list.append(np.nan)
                feat_values.append(tmp_list)
                continue

            # Compute features related to ionization/affinity.
            cation_fraction = anion_fraction = cation_ie_sum = anion_ea_sum = \
            mean_cation_en = mean_anion_en = 0.0
            for i,elem in enumerate(elems):
                if tmp_charges[i] < 0:
                    anion_fraction += fracs[i]
                    mean_anion_en +=  en[elem] * fracs[i]
                    anion_ea_sum -= tmp_charges[i] * ea[elem]* fracs[i]
                else:
                    cation_fraction += fracs[i]
                    mean_cation_en += en[elem] * fracs[i]
                    cation_ie_sum += sum(ie[elem][c] * fracs[i] for c in
                                         range(int(tmp_charges[i])))

            mean_anion_en /= anion_fraction
            mean_cation_en /= cation_fraction
            anion_ea_sum /= anion_fraction
            cation_ie_sum /= cation_fraction

            tmp_list.append(cation_ie_sum)
            tmp_list.append(anion_ea_sum)
            tmp_list.append(mean_anion_en - mean_cation_en)

            feat_values.append(tmp_list)

        # Issue warning to user about missing data here if it exists.
        if len(missing_data) > 0:
            sys.stderr.write("WARNING: Missing ionization energy data for:\n")
            for elem in missing_data:
                sys.stderr.write("\t" + LookUpData.element_names[elem] + ":")
                for state in missing_data[elem]:
                    sys.stderr.write(" +" + str(state))
                sys.stderr.write("\n")

        features = pd.DataFrame(feat_values, columns=feat_headers)

        return features


import unittest
import numpy as np

from cheml.chem import BagofBonds


# Oxygen, Hydrogen, Hydrogen
n = np.array([8, 1, 1])
c = np.array([[1.464, 0.707, 1.056],
             [0.878, 1.218, 0.498],
             [2.319, 1.126, 0.952]])

nc = n.reshape((3,1))
nc = np.append(nc,c,1)

class TestBagofBonds(unittest.TestCase):
    def test_h2o(self):
        bob = BagofBonds(const=1.0)
        h2o = bob.represent(np.array([nc]))

        self.assertEqual(h2o.shape, (1, 6))
        a = np.array([[  0.66066557,   0.5       ,   0.5       ,   8.3593106 ,
          8.35237809,  73.51669472]])
        self.assertAlmostEqual(a[0][0], h2o.values[0][0])
        self.assertAlmostEqual(a[0][1], h2o.values[0][1])
        self.assertAlmostEqual(a[0][-1], h2o.values[0][-1])




if __name__== '__main__':
    unittest.main()




import tensorflow as tf
from ...utils import choice, chunk
from ...utils import check_input

class mlp_classification(object):
    def __init__(self,nneurons, act_funcs, cost='mse', optimizer='GradientDescentOptimizer', validation_size=0.2,
                 learning_rate=0.001, training_epochs=1000, batch_size = 100, display_step = 10 ):
        """
        multilayer perceptron with tensorflow

        :param nneurons: list of integers
            describing how many neurons there are in each hidden layer
        :param act_funcs: list of activation functions or just one function (string)
            should be same length as nneurons if a list
        :param cost: string (default = 'mse')
            List of available cost functions:
                - 'mse': mean squared error
                - 'scel': softmax cross_entropy with logits
        :param optimizer: string (default = 'GradientDescentOptimizer')
            List of available optimizers:
                - GradientDescentOptimizer
                - AdadeltaOptimizer
                - AdagradOptimizer
                - AdagradDAOptimizer
                - MomentumOptimizer
                - AdamOptimizer
                - FtrlOptimizer
                - ProximalGradientDescentOptimizer
                - ProximalAdagradOptimizer
                - RMSPropOptimizer
        :param validation_size: float between zero and one, optional (default = 0.2)
            size of data to be selected randomly for validation
        :param learning_rate:
        :param training_epochs:
        :param batch_size:
        :param display_step:
        """
        self.nneurons = nneurons
        self.act_funcs = act_funcs
        self.cost = cost
        self.optimizer = optimizer
        self.validation_size = validation_size
        self.learning_rate = learning_rate
        self.training_epochs = training_epochs
        self.batch_size = batch_size
        self.display_step = display_step

    def train(self, X, Y):
        """
        make tensorflow variables and fit all the parameters. Then train the network parameters.

        :param X: pandas dataframe or numpy array
            input training data
        :param Y: pandas dataframe or numpy array
            output training data
        :return training accuracy: float
            validation accuracy: float
        """
        X_train, _ = check_input(X, 'Training input', format_out='ar')
        Y_train, _ = check_input(Y, 'Training output', n0=X_train.shape[0], format_out='ar')
        X_train, X_test, Y_train, Y_test = choice(X_train, Y_train, n=self.validation_size)
        X_test, _ = check_input(X_test, 'Testing input', n1=X_train.shape[1], format_out='ar')
        Y_test, _ = check_input(Y_test, 'Testing output', n0=X_test.shape[0], n1=Y_train.shape[1], format_out='ar')
        self.n_features = X_train.shape[1]
        self.n_outputs = Y_train.shape[1]
        N = X_train.shape[0]

        # tf Graph input
        x = tf.placeholder("float", [None, self.n_features])
        y = tf.placeholder("float", [None, self.n_outputs])
        self._initialize_weights()
        self.mapped_act_funcs = self._act_funcs_from_string(self.act_funcs)
        y_pred = self._output(x)
        cost = self._loss_function(y_pred,y)
        optimizer = self._optimizer(cost)
        init = tf.initialize_all_variables()
        # Training cycle
        self.sess = tf.Session()
        self.sess.run(init)
        for epoch in range(self.training_epochs):
            avg_cost = 0.
            total_batch = N // self.batch_size
            # Loop over all batches
            it = chunk(range(len(X_train)),total_batch,X_train,Y_train)
            for i in xrange(total_batch):
                batch_x, batch_y = it.next()
                # Run optimization op (backprop) and cost op (to get loss value)
                _, c = self.sess.run([optimizer, cost], feed_dict={x: batch_x,
                                                              y: batch_y})
                # Compute average loss
                avg_cost += c / total_batch
            # Display logs per epoch step
            if epoch % self.display_step == 0:
                print("Epoch:", '%04d' % (epoch + 1), "cost=", \
                      "{:.9f}".format(avg_cost))

        self.nnet = {'weights':self.sess.run(self.weights), 'biases':self.sess.run(self.biases), 'act_func':self.act_funcs}

        correct_prediction = tf.equal(tf.argmax(y_pred, 1), tf.argmax(y, 1))
        # Calculate accuracy
        accuracy = tf.reduce_mean(tf.cast(correct_prediction, "float"))
        training_accuracy = self.score(X_train, Y_train)
        validation_accuracy = self.score(X_test, Y_test)
        print '______________________'
        print "Training Accuracy:", training_accuracy
        print "Validation Accuracy:", validation_accuracy

        return training_accuracy, validation_accuracy

    def predict(self,X):
        """
        predict labels for the input data
        :param X: 2d array-like
            input data
        :return: predictions
            predicted labels
        """
        x = tf.placeholder("float", [None, self.n_features])
        y_pred = self._output(x)
        y_pred = self.sess.run(y_pred,feed_dict={x:X})
        return y_pred

    def score(self, X, Y):
        """
        returns fraction of correctly classified samples in the given data (X).

        :param X: 2d array-like
            input data with same number of features as the input training data
        :param Y: 1d/2d array-like
            labels with same dimension as input training labels
        :return: score: 1d array-like
            the prediction accuracy of model
        """
        x = tf.placeholder("float", [None, self.n_features])
        y = tf.placeholder("float", [None, self.n_outputs])
        y_pred = self._output(x)
        # if task=='regression':
        #     accuracy = tf.reduce_mean(tf.abs(tf.subtract(y,y_pred)))
        # elif task=='classification':
        correct_prediction = tf.equal(tf.argmax(y_pred, 1), tf.argmax(y, 1))
        accuracy = tf.reduce_mean(tf.cast(correct_prediction, "float"))
        score = accuracy.eval(session=self.sess, feed_dict={x: X, y: Y})
        return score

    # def call_predict(self, nnet, X, Y=None):
    #     """
    #
    #     :param nnet: dict, default(None)
    #         parameters of a trained neural network: keys required weights, biases and activation functions
    #     :param X:
    #     :param Y:
    #     :return:
    #     """

    def _initialize_weights(self):
        self.weights = []
        self.biases = []
        complete_nneurons = [self.n_features] + self.nneurons + [self.n_outputs]
        for prev_nneurons, nneurons in zip(complete_nneurons[:-1], complete_nneurons[1:]):
            self.weights.append(tf.Variable(tf.random_normal([prev_nneurons, nneurons])))
            self.biases.append(tf.Variable(tf.random_normal([nneurons])))

    def _act_funcs_from_string(self, act_funcs):
        """
        translate activation functions input from user
        """
        act_func_dict = {'relu': tf.nn.relu,
                         'relu6': tf.nn.relu6,
                         # 'crelu': tf.nn.crelu,
                         'elu': tf.nn.elu,
                         'softplus': tf.nn.softplus,
                         'softsign': tf.nn.softsign,
                         'sigmoid': tf.sigmoid,
                         'tanh': tf.tanh}
        if isinstance(act_funcs, list):
            if len(act_funcs) != len(self.nneurons):
                msg = 'List of activation function is of different length from list of neuron numbers'
                raise Exception(msg)
            else:
                return map(lambda x: act_func_dict[x], act_funcs)
        else:
            return [act_func_dict[act_funcs]] * (len(self.nneurons))

    def _output(self,x):
        hi = x
        for i in xrange(len(self.nneurons)):
            hi = self.mapped_act_funcs[i](tf.add(tf.matmul(hi, self.weights[i]), self.biases[i]))
        out_layer = tf.add(tf.matmul(hi, self.weights[-1]), self.biases[-1])
        return out_layer

    def _loss_function(self,y_pred,y):
        if self.cost == 'scel':
            cost = tf.reduce_mean(tf.nn.softmax_cross_entropy_with_logits(y_pred, y))
        elif self.cost == 'mse':
            cost = tf.reduce_mean(tf.square(tf.sub(y_pred, y)))
        return cost

    def _optimizer(self,cost):
        if self.optimizer == 'GradientDescentOptimizer':
            optimizer = tf.train.GradientDescentOptimizer(learning_rate=self.learning_rate).minimize(cost)
        elif self.optimizer == 'AdadeltaOptimizer':
            optimizer = tf.train.AdadeltaOptimizer(learning_rate=self.learning_rate).minimize(cost)
        elif self.optimizer == 'AdagradOptimizer':
            optimizer = tf.train.AdagradOptimizer(learning_rate=self.learning_rate).minimize(cost)
        elif self.optimizer == 'AdagradDAOptimizer':
            optimizer = tf.train.AdagradDAOptimizer(learning_rate=self.learning_rate).minimize(cost)
        elif self.optimizer == 'MomentumOptimizer':
            optimizer = tf.train.MomentumOptimizer(learning_rate=self.learning_rate).minimize(cost)
        elif self.optimizer == 'AdamOptimizer':
            optimizer = tf.train.AdamOptimizer(learning_rate=self.learning_rate).minimize(cost)
        elif self.optimizer == 'FtrlOptimizer':
            optimizer = tf.train.FtrlOptimizer(learning_rate=self.learning_rate).minimize(cost)
        elif self.optimizer == 'ProximalGradientDescentOptimizer':
            optimizer = tf.train.ProximalGradientDescentOptimizer(learning_rate=self.learning_rate).minimize(cost)
        elif self.optimizer == 'ProximalAdagradOptimizer':
            optimizer = tf.train.ProximalAdagradOptimizer(learning_rate=self.learning_rate).minimize(cost)
        elif self.optimizer == 'RMSPropOptimizer':
            optimizer = tf.train.RMSPropOptimizer(learning_rate=self.learning_rate).minimize(cost)
        return optimizer

    def test(self):
        from tensorflow.examples.tutorials.mnist import input_data
        mnist = input_data.read_data_sets("/tmp/data/", one_hot=True)
        X=mnist.train.images
        Y=mnist.train.labels
        _, _ = self.train(X,Y)
        _ = self.predict(mnist.test.images)
        X_test = mnist.test.images
        Y_test = mnist.test.labels
        score = self.score(X_test, Y_test)
        return score

class mlp_regression(object):
    def __init__(self,nneurons, act_funcs, cost='mse', optimizer='GradientDescentOptimizer', validation_size=0.2,
                 learning_rate=0.001, training_epochs=1000, batch_size = 100, display_step = 10 ):
        """
        multilayer perceptron with tensorflow

        :param nneurons: list of integers
            describing how many neurons there are in each hidden layer
        :param act_funcs: list of activation functions or just one function (string)
            should be same length as nneurons if a list
        :param cost: string (default = 'mse')
            List of available cost functions:
                - 'mse': mean squared error
                - 'scel': softmax cross_entropy with logits
        :param optimizer: string (default = 'GradientDescentOptimizer')
            List of available optimizers:
                - GradientDescentOptimizer
                - AdadeltaOptimizer
                - AdagradOptimizer
                - AdagradDAOptimizer
                - MomentumOptimizer
                - AdamOptimizer
                - FtrlOptimizer
                - ProximalGradientDescentOptimizer
                - ProximalAdagradOptimizer
                - RMSPropOptimizer
        :param validation_size: float between zero and one, optional (default = 0.2)
            size of data to be selected randomly for validation
        :param learning_rate:
        :param training_epochs:
        :param batch_size:
        :param display_step:
        """
        self.nneurons = nneurons
        self.act_funcs = act_funcs
        self.cost = cost
        self.optimizer = optimizer
        self.validation_size = validation_size
        self.learning_rate = learning_rate
        self.training_epochs = training_epochs
        self.batch_size = batch_size
        self.display_step = display_step

    def train(self, X, Y):
        """
        make tensorflow variables and fit all the parameters. Then train the network parameters.

        :param X: pandas dataframe or numpy array
            input training data
        :param Y: pandas dataframe or numpy array
            output training data
        :return training accuracy: float
            validation accuracy: float
        """
        X_train, _ = check_input(X, 'Training input', format_out='ar')
        Y_train, _ = check_input(Y, 'Training output', n0=X_train.shape[0], format_out='ar')
        X_train, X_test, Y_train, Y_test = choice(X_train, Y_train, n=self.validation_size)
        X_test, _ = check_input(X_test, 'Testing input', n1=X_train.shape[1], format_out='ar')
        Y_test, _ = check_input(Y_test, 'Testing output', n0=X_test.shape[0], n1=Y_train.shape[1], format_out='ar')
        self.n_features = X_train.shape[1]
        self.n_outputs = Y_train.shape[1]
        N = X_train.shape[0]

        # tf Graph input
        x = tf.placeholder("float", [None, self.n_features])
        y = tf.placeholder("float", [None, self.n_outputs])
        self._initialize_weights()
        self.mapped_act_funcs = self._act_funcs_from_string(self.act_funcs)
        y_pred = self._output(x)
        cost = self._loss_function(y_pred,y)
        optimizer = self._optimizer(cost)
        init = tf.initialize_all_variables()
        # Training cycle
        self.sess = tf.Session()
        self.sess.run(init)
        for epoch in range(self.training_epochs):
            avg_cost = 0.
            total_batch = N // self.batch_size
            # Loop over all batches
            it = chunk(range(len(X_train)),total_batch,X_train,Y_train)
            for i in xrange(total_batch):
                batch_x, batch_y = it.next()
                # Run optimization op (backprop) and cost op (to get loss value)
                _, c = self.sess.run([optimizer, cost], feed_dict={x: batch_x,
                                                              y: batch_y})
                # Compute average loss
                avg_cost += c / total_batch
            # Display logs per epoch step
            if epoch % self.display_step == 0:
                print("Epoch:", '%04d' % (epoch + 1), "cost=", \
                      "{:.9f}".format(avg_cost))

        self.nnet = {'weights':self.sess.run(self.weights), 'biases':self.sess.run(self.biases), 'act_func':self.act_funcs}

        correct_prediction = tf.equal(tf.argmax(y_pred, 1), tf.argmax(y, 1))
        # Calculate accuracy
        accuracy = tf.reduce_mean(tf.cast(correct_prediction, "float"))
        training_accuracy = self.score(X_train, Y_train)
        validation_accuracy = self.score(X_test, Y_test)
        print '______________________'
        print "Training Accuracy:", training_accuracy
        print "Validation Accuracy:", validation_accuracy

        return training_accuracy, validation_accuracy

    def predict(self,X):
        """
        predict labels for the input data
        :param X: 2d array-like
            input data
        :return: predictions
            predicted labels
        """
        x = tf.placeholder("float", [None, self.n_features])
        y_pred = self._output(x)
        y_pred = self.sess.run(y_pred,feed_dict={x:X})
        return y_pred

    def score(self, X, Y):
        """
        returns mean absolute error on the given data (X) and labels (Y)

        :param X: 2d array-like
            input data with same number of features as the input training data
        :param Y: 1d/2d array-like
            labels with same dimension as input training labels
        :return: score: 1d array-like
            the prediction accuracy of model
        """
        x = tf.placeholder("float", [None, self.n_features])
        y = tf.placeholder("float", [None, self.n_outputs])
        y_pred = self._output(x)
        accuracy = accuracy = tf.reduce_mean(tf.abs(tf.subtract(y,y_pred)))
        score = accuracy.eval(session=self.sess, feed_dict={x: X, y: Y})
        return score

    def _initialize_weights(self):
        self.weights = []
        self.biases = []
        complete_nneurons = [self.n_features] + self.nneurons + [self.n_outputs]
        for prev_nneurons, nneurons in zip(complete_nneurons[:-1], complete_nneurons[1:]):
            self.weights.append(tf.Variable(tf.random_normal([prev_nneurons, nneurons])))
            self.biases.append(tf.Variable(tf.random_normal([nneurons])))

    def _act_funcs_from_string(self, act_funcs):
        """
        translate activation functions input from user
        """
        act_func_dict = {'relu': tf.nn.relu,
                         'relu6': tf.nn.relu6,
                         # 'crelu': tf.nn.crelu,
                         'elu': tf.nn.elu,
                         'softplus': tf.nn.softplus,
                         'softsign': tf.nn.softsign,
                         'sigmoid': tf.sigmoid,
                         'tanh': tf.tanh}
        if isinstance(act_funcs, list):
            if len(act_funcs) != len(self.nneurons):
                msg = 'List of activation function is of different length from list of neuron numbers'
                raise Exception(msg)
            else:
                return map(lambda x: act_func_dict[x], act_funcs)
        else:
            return [act_func_dict[act_funcs]] * (len(self.nneurons))

    def _output(self,x):
        hi = x
        for i in xrange(len(self.nneurons)):
            hi = self.mapped_act_funcs[i](tf.add(tf.matmul(hi, self.weights[i]), self.biases[i]))
        out_layer = tf.add(tf.matmul(hi, self.weights[-1]), self.biases[-1])
        return out_layer

    def _loss_function(self,y_pred,y):
        if self.cost == 'scel':
            cost = tf.reduce_mean(tf.nn.softmax_cross_entropy_with_logits(y_pred, y))
        elif self.cost == 'mse':
            cost = tf.reduce_mean(tf.square(tf.sub(y_pred, y)))
        return cost

    def _optimizer(self,cost):
        if self.optimizer == 'GradientDescentOptimizer':
            optimizer = tf.train.GradientDescentOptimizer(learning_rate=self.learning_rate).minimize(cost)
        elif self.optimizer == 'AdadeltaOptimizer':
            optimizer = tf.train.AdadeltaOptimizer(learning_rate=self.learning_rate).minimize(cost)
        elif self.optimizer == 'AdagradOptimizer':
            optimizer = tf.train.AdagradOptimizer(learning_rate=self.learning_rate).minimize(cost)
        elif self.optimizer == 'AdagradDAOptimizer':
            optimizer = tf.train.AdagradDAOptimizer(learning_rate=self.learning_rate).minimize(cost)
        elif self.optimizer == 'MomentumOptimizer':
            optimizer = tf.train.MomentumOptimizer(learning_rate=self.learning_rate).minimize(cost)
        elif self.optimizer == 'AdamOptimizer':
            optimizer = tf.train.AdamOptimizer(learning_rate=self.learning_rate).minimize(cost)
        elif self.optimizer == 'FtrlOptimizer':
            optimizer = tf.train.FtrlOptimizer(learning_rate=self.learning_rate).minimize(cost)
        elif self.optimizer == 'ProximalGradientDescentOptimizer':
            optimizer = tf.train.ProximalGradientDescentOptimizer(learning_rate=self.learning_rate).minimize(cost)
        elif self.optimizer == 'ProximalAdagradOptimizer':
            optimizer = tf.train.ProximalAdagradOptimizer(learning_rate=self.learning_rate).minimize(cost)
        elif self.optimizer == 'RMSPropOptimizer':
            optimizer = tf.train.RMSPropOptimizer(learning_rate=self.learning_rate).minimize(cost)
        return optimizer

    def test(self):
        from tensorflow.examples.tutorials.mnist import input_data
        mnist = input_data.read_data_sets("/tmp/data/", one_hot=True)
        X=mnist.train.images
        Y=mnist.train.labels
        _, _ = self.train(X,Y)
        _ = self.predict(mnist.test.images)
        X_test = mnist.test.images
        Y_test = mnist.test.labels
        score = self.score(X_test, Y_test)
        return score



import os
import sys
import pandas as pd
import numpy as np
import warnings
import fnmatch

from ..utils.utilities import std_datetime_str


class Split(object):
    """
    split data frame by columns

    Parameters
    ----------
    select: integer or list (default = 1)
        integer: number of columns to be selected from the first of data as first data frame (X1)
        list: list of headers to be selected as first data frame (X1)

    Returns
    -------
    two pandas dataframes: X1 and X2
    """
    def __init__(self,selection=1):
        self.selection = selection

    def fit(self,X):
        """
        fit the split task to the input data frame

        :param X:  pandas data frame
        original pandas data frame
        :return: two pandas data frame: X1 and X2
        """
        if not isinstance(X,pd.DataFrame):
            msg = 'X must be a pandas dataframe'
            raise TypeError(msg)
        if isinstance(self.selection,list):
            X1 = X.loc[:,self.selection]
            X2 = X.drop(self.selection,axis=1)
        elif isinstance(self.selection,int):
            if self.selection >= X.shape[1]:
                msg = 'The first output data frame is empty, because passed a bigger number than actual number of columns'
                warnings.warn(msg)
            X1 = X.iloc[:,:self.selection]
            X2 = X.iloc[:,self.selection:]
        else:
            msg = "selection parameter must ba a list or an integer"
            raise TypeError(msg)
        return X1, X2

class XYZreader(object):
    """ (XYZreader)
    Read molecules' geometry (cartesian coordinates) from one or more XYZ files.

    Parameters
    ----------
    path_pattern: string or list of string
        A pattern or a list of patterns. Each pattern consists of file name and path.
        The pattern can contain any special character in the following list:

            *       : matches everything (zero or more characters)
            ?       : matches any single character
            [seq]   : matches any character in seq
            [!seq]  : matches any character not in seq
            /       : filename seperator

        Note: seq can indicate a range of characters by giving two characters and separating them by a '-'.
              However, the range is limited to a single character. That's why this parameter also accept a list of
              patterns for different length of characters. For example, range(1,30) = '[1-9]' and '[1-9][0-9]'

        The pattern matching is utilizes fnmatch library- Unix filename pattern matching.
        Note: The pattern must include the file format at the end. The only acceptable extension is '.xyz'.

    path_root: string, optional (default = None)
        fixed (with no special character) part of the path.
        If path is None, this function tries to open the file as a single
        file (without any pattern matching). It is a fixed path to subdirectories
        or files. Therefore, none of the above special characters except '/'
        can be used.
        If not None, it determines the path that this function walk through
        and look for every file that matches the pattern. To start
        walking from the curent directory, the path should be '.'

    Z: dictionary, optional (default = {'Ru': 44.0, 'Re': 75.0, 'Rf': 104.0, 'Rg': 111.0, 'Ra': 88.0, 'Rb': 37.0, 'Rn': 86.0, 'Rh': 45.0, 'Be': 4.0, 'Ba': 56.0, 'Bh': 107.0, 'Bi': 83.0, 'Bk': 97.0, 'Br': 35.0, 'H': 1.0, 'P': 15.0, 'Os': 76.0, 'Es': 99.0, 'Hg': 80.0, 'Ge': 32.0, 'Gd': 64.0, 'Ga': 31.0, 'Pr': 59.0, 'Pt': 78.0, 'Pu': 94.0, 'C': 6.0, 'Pb': 82.0, 'Pa': 91.0, 'Pd': 46.0, 'Cd': 48.0, 'Po': 84.0, 'Pm': 61.0, 'Hs': 108.0, 'Uup': 115.0, 'Uus': 117.0, 'Uuo': 118.0, 'Ho': 67.0, 'Hf': 72.0, 'K': 19.0, 'He': 2.0, 'Md': 101.0, 'Mg': 12.0, 'Mo': 42.0, 'Mn': 25.0, 'O': 8.0, 'Mt': 109.0, 'S': 16.0, 'W': 74.0, 'Zn': 30.0, 'Eu': 63.0, 'Zr': 40.0, 'Er': 68.0, 'Ni': 28.0, 'No': 102.0, 'Na': 11.0, 'Nb': 41.0, 'Nd': 60.0, 'Ne': 10.0, 'Np': 93.0, 'Fr': 87.0, 'Fe': 26.0, 'Fl': 114.0, 'Fm': 100.0, 'B': 5.0, 'F': 9.0, 'Sr': 38.0, 'N': 7.0, 'Kr': 36.0, 'Si': 14.0, 'Sn': 50.0, 'Sm': 62.0, 'V': 23.0, 'Sc': 21.0, 'Sb': 51.0, 'Sg': 106.0, 'Se': 34.0, 'Co': 27.0, 'Cn': 112.0, 'Cm': 96.0, 'Cl': 17.0, 'Ca': 20.0, 'Cf': 98.0, 'Ce': 58.0, 'Xe': 54.0, 'Lu': 71.0, 'Cs': 55.0, 'Cr': 24.0, 'Cu': 29.0, 'La': 57.0, 'Li': 3.0, 'Lv': 116.0, 'Tl': 81.0, 'Tm': 69.0, 'Lr': 103.0, 'Th': 90.0, 'Ti': 22.0, 'Te': 52.0, 'Tb': 65.0, 'Tc': 43.0, 'Ta': 73.0, 'Yb': 70.0, 'Db': 105.0, 'Dy': 66.0, 'Ds': 110.0, 'I': 53.0, 'U': 92.0, 'Y': 39.0, 'Ac': 89.0, 'Ag': 47.0, 'Uut': 113.0, 'Ir': 77.0, 'Am': 95.0, 'Al': 13.0, 'As': 33.0, 'Ar': 18.0, 'Au': 79.0, 'At': 85.0, 'In': 49.0})
        A dictionary of nuclear charges with respect to the chemical symbols of all atom types in the xyz files.

    reader: string, optional (default = 'auto')
        Available options : 'auto' and 'manual'
        If 'auto', the openbabel readstring function creat the molecule object.
        The type of files for openbabel class has been set to 'xyz', thus the format
        of the file should also follow a typical xyz format. However, with 'manual'
        reader you can skip some lines from top or bottom of xyz files.

    skip_lines: list of two integers, optional (default = [2,0])
        Number of lines to skip (int) from top and bottom of the xyz files, respectively.
        Based on the original xyz format only 2 lines from top and zero from
        bottom of a file can be skipped. Thus, number of atoms in the first line
        can also be ignored.
        Only available for 'manual' reader.

    Attributes
    ----------
    max_n_atoms: integer
        Maximum number of atoms in one molecule.
        This can be useful if you want to set this parameter in the feature representation methods,
        e.g. Coulomb_Matrix.

    Examples
    --------
        (1)
        path_pattern: 'Mydir/1f/1_opt.xyz'
        path_root: None
        >>> one file will be read: 'Mydir/1f/1_opt.xyz'

        (2)
        path_pattern: '[1,2,3,4]?/*_opt.xyz'
        path_root: 'Mydir'
        >>> sample files to be read: 'Mydir/1f/1_opt.xyz', 'Mydir/2c/2_opt.xyz', ...

        (3)
        path_pattern: '[!1,2]?/*_opt.xyz'
        path_root: '.'
        >>> sample files to be read: './3f/3_opt.xyz', 'Mydir/7c/7_opt.xyz', ...

        (4)
        path_pattern: '*['f','c']/*_opt.xyz'
        path_root: 'Mydir'
        >>> sample files to be read: 'Mydir/1f/1_opt.xyz', 'Mydir/2c/2_opt.xyz', ...

        (5)
        path_pattern: ['[2-5]_opt.xyz', '[1-9][2-5]_opt.xyz']
        path_root: 'Mydir/all'
        >>> sample files to be read: 'Mydir/all/1f/1_opt.xyz', 'Mydir/all/2c/2_opt.xyz', ...
    """
    def __init__(self, path_pattern, path_root=None, Z = {'Ru': 44.0, 'Re': 75.0, 'Rf': 104.0, 'Rg': 111.0, 'Ra': 88.0, 'Rb': 37.0, 'Rn': 86.0, 'Rh': 45.0, 'Be': 4.0, 'Ba': 56.0, 'Bh': 107.0, 'Bi': 83.0, 'Bk': 97.0, 'Br': 35.0, 'H': 1.0, 'P': 15.0, 'Os': 76.0, 'Es': 99.0, 'Hg': 80.0, 'Ge': 32.0, 'Gd': 64.0, 'Ga': 31.0, 'Pr': 59.0, 'Pt': 78.0, 'Pu': 94.0, 'C': 6.0, 'Pb': 82.0, 'Pa': 91.0, 'Pd': 46.0, 'Cd': 48.0, 'Po': 84.0, 'Pm': 61.0, 'Hs': 108.0, 'Uup': 115.0, 'Uus': 117.0, 'Uuo': 118.0, 'Ho': 67.0, 'Hf': 72.0, 'K': 19.0, 'He': 2.0, 'Md': 101.0, 'Mg': 12.0, 'Mo': 42.0, 'Mn': 25.0, 'O': 8.0, 'Mt': 109.0, 'S': 16.0, 'W': 74.0, 'Zn': 30.0, 'Eu': 63.0, 'Zr': 40.0, 'Er': 68.0, 'Ni': 28.0, 'No': 102.0, 'Na': 11.0, 'Nb': 41.0, 'Nd': 60.0, 'Ne': 10.0, 'Np': 93.0, 'Fr': 87.0, 'Fe': 26.0, 'Fl': 114.0, 'Fm': 100.0, 'B': 5.0, 'F': 9.0, 'Sr': 38.0, 'N': 7.0, 'Kr': 36.0, 'Si': 14.0, 'Sn': 50.0, 'Sm': 62.0, 'V': 23.0, 'Sc': 21.0, 'Sb': 51.0, 'Sg': 106.0, 'Se': 34.0, 'Co': 27.0, 'Cn': 112.0, 'Cm': 96.0, 'Cl': 17.0, 'Ca': 20.0, 'Cf': 98.0, 'Ce': 58.0, 'Xe': 54.0, 'Lu': 71.0, 'Cs': 55.0, 'Cr': 24.0, 'Cu': 29.0, 'La': 57.0, 'Li': 3.0, 'Lv': 116.0, 'Tl': 81.0, 'Tm': 69.0, 'Lr': 103.0, 'Th': 90.0, 'Ti': 22.0, 'Te': 52.0, 'Tb': 65.0, 'Tc': 43.0, 'Ta': 73.0, 'Yb': 70.0, 'Db': 105.0, 'Dy': 66.0, 'Ds': 110.0, 'I': 53.0, 'U': 92.0, 'Y': 39.0, 'Ac': 89.0, 'Ag': 47.0, 'Uut': 113.0, 'Ir': 77.0, 'Am': 95.0, 'Al': 13.0, 'As': 33.0, 'Ar': 18.0, 'Au': 79.0, 'At': 85.0, 'In': 49.0},
              reader='auto', skip_lines=[2, 0], path_only = False):
        self.path_pattern = path_pattern
        self.path_root = path_root
        self.Z = Z
        self.reader = reader
        self.skip_lines = skip_lines
        self.path_only = path_only

    def __file_reader(self, filename):
        if self.reader == 'auto':
            # sys.path.insert(0, "/user/m27/pkg/openbabel/2.3.2/lib")
            import pybel
            # import openbabel
            mol = open(filename, 'r').read()
            mol = pybel.readstring("xyz", mol)
            molecule = [(a.OBAtom.GetAtomicNum(), a.OBAtom.x(), a.OBAtom.y(), a.OBAtom.z()) for a in mol.atoms]
            return np.array(molecule)
        elif self.reader == 'manual':
            mol = open(filename, 'r').readlines()
            if len(mol) == 0:
                return np.array([])
            mol = mol[self.skip_lines[0]:len(mol) - self.skip_lines[1]]
            molecule = []
            for atom in mol:
                atom = atom.replace('\t', ' ')
                atom = atom.strip().split(' ')
                atom = list(filter(lambda x: x != '', atom))
                molecule.append([self.Z[atom[0]], float(atom[1]), float(atom[2]), float(atom[3])])
            return np.array(molecule)

    def read(self):
        """
        read the XYZ files based on the path_pattern and path_root parameters and create the output ductionary.

        Return
        ------
        molecules: dictionary
            dictionary of molecules with ['mol', 'file'] keys
            The value of 'file' is the name and path of the file that was read to create the molecule.
            The value of 'mol' is the numpy array of 4 values for each atom in the molecule. The four values are
             nuclear charge, X-coordinate, Y-coordinate, and Z-coordinate, respectively.
        """
        if isinstance(self.path_pattern, str):
            self.path_pattern = [self.path_pattern]
        #Todo: change the molecules to a list of dictionaries
        molecules = {}
        it = 0
        max_nAtoms = 1
        for pattern in self.path_pattern:
            file_name, file_extension = os.path.splitext(pattern)
            if file_extension == '':
                msg = 'file extension not indicated'
                raise ValueError(msg)
            elif file_extension != '.xyz':
                msg = "file extension '%s' not available - xyz is the only acceptable extension" % file_extension
                raise ValueError(msg)

            if self.path_root:
                for root, directories, filenames in os.walk(self.path_root):
                    file_path = [os.path.join(root, filename) for filename in filenames]
                    for fn in fnmatch.filter(file_path, os.path.join(self.path_root, pattern)):
                        if not self.path_only:
                            mol = self.__file_reader(fn)
                            max_nAtoms = max(max_nAtoms, len(mol))
                        else:
                            mol = None
                            max_nAtoms = max(max_nAtoms, 0)
                        it+=1
                        molecules[it] = {'file':fn, 'mol': mol}
            else:
                if not self.path_only:
                    mol = self.__file_reader(file_name)
                    max_nAtoms = max(max_nAtoms, len(mol))
                else:
                    mol = None
                    max_nAtoms = max(max_nAtoms, 0)
                it += 1
                molecules[it] = {'file': file_name, 'mol': mol}
        self.max_n_atoms = max_nAtoms
        return molecules

class ConvertFile(object):
    """(ConvertFile)
    Specify file path, 'from_format' and 'to_format' to convert a file form 'from_format' to 'to_format'
    using openbabel (https://openbabel.org/wiki/Babel).

    Parameters:
    ----------
    file_path : string or dictionary
        string or dictionary containing paths of files that need to be converted

    from_format: string
        String of letters that specify the format of the file that needs to be converted.
        This will be checked against 'file_path' that is provided by the user.
        If the file_path does not contain 'from_format' an error message will be raised.
        List of possible 'from_format's are on https://openbabel.org/wiki/Babel

    to_format: string
        String of letters that specify the target file format or the desired format.
        An openbabel command is generated which converts the files specified by 'file_path' to the target format.
        List of possible 'to_format's are on https://openbabel.org/wiki/Babel

    Returns:
    ------
    converted_file_paths: dictionary
        dictionary of converted file paths, in the same format of XYZreader output: {index:{'file':""}}

    Examples:
    --------
    >>> from cheml.datasets import load_xyz_polarizability
    >>> coordinates,polarizability = load_xyz_polarizability()
    >>> coordinates
    {1: {'file': 'cheml/datasets/data/organic_xyz/1_opt.xyz', ...
    >>> from cheml.initialization import ConvertFile
    >>> model = ConvertFile(file_path=coordinates,from_format='xyz',to_format='cml')
    >>> converted_file_paths = model.convert()
    {1: {'file': 'cheml/datasets/data/organic_xyz/1_opt.cml'}, 2: ...

    """

    def __init__(self,file_path,from_format,to_format):
        self.file_path=file_path
        self.from_format=from_format
        self.to_format=to_format

    def convert(self):
        converted_file_paths={}
        if isinstance(self.file_path,str):
            if not self.from_format == self.file_path[-len(self.from_format):]:
                msg = 'file format is not the same as from_format'
                raise ValueError(msg)

            else:
                ob_from_format = '-i' + self.from_format
                ob_to_format = '-o' + self.to_format
                path=self.file_path[:self.file_path.rfind('.')+1]
                command='babel ' + ob_from_format + ' ' + self.file_path + ' ' + ob_to_format + ' ' + path+self.to_format
                print command
                os.system(command)
                converted_file_paths[1] = {'file':path+self.to_format}

        elif isinstance(self.file_path,dict):
            for it in range(1,len(self.file_path)+1):
                fpath=self.file_path[it]['file']
                if not fpath[-len(self.from_format):] == self.from_format:
                    msg='file format is not the same as from_format'
                    raise ValueError(msg)
                else:
                    ob_from_format = '-i' + self.from_format
                    ob_to_format = '-o' + self.to_format
                    path=fpath[:fpath.rfind('.')+1]
                    command = 'babel ' + ob_from_format + ' ' + fpath + ' ' + ob_to_format + ' ' + path + self.to_format
                    os.system(command)
                    converted_file_paths[it] = {'file':path+self.to_format}

        return converted_file_paths

class SaveFile(object):
    """
    Write DataFrame to a comma-seprated values(csv) file
    :param: output_directory: string, the output directory to save output files

    """
    def __init__(self, filename, output_directory = None, record_time = False, format ='csv',
                 index = False, header = True):
        self.filename = filename
        self.record_time = record_time
        self.output_directory = output_directory
        self.format = format
        self.index = index
        self.header = header

    def fit(self, X, main_directory='.'):
        """
        Write DataFrame to a comma-seprated values (csv) file

        Parameters
        ----------
        X: pandas DataFrame
        main_directory: string
            if there is any main directory for entire cheml project
        """
        if not isinstance(X, pd.DataFrame):
            msg = 'X must be a pandas dataframe'
            raise TypeError(msg)

        if self.output_directory:
            self.output_directory = os.path.join(main_directory, self.output_directory)
            if not os.path.exists(self.output_directory):
                os.makedirs(self.output_directory)
            if self.record_time:
                self.file_path = '%s/%s_%s.%s'%(self.output_directory, self.filename,std_datetime_str(),self.format)
                X.to_csv(self.file_path, index=self.index, header = self.header)
            else:
                self.file_path = '%s/%s.%s' % (self.output_directory, self.filename,self.format)
                X.to_csv(self.file_path, index=self.index, header = self.header)
        else:
            if self.record_time:
                self.file_path = '%s/%s_%s.%s'%(main_directory,self.filename,std_datetime_str(),self.format)
                X.to_csv(self.file_path, index=self.index, header = self.header)
            else:
                self.file_path = '%s/%s.%s' %(main_directory,self.filename,self.format)
                X.to_csv(self.file_path, index=self.index, header = self.header)

class StoreFile(object):
    """
    Write everything in to a text file
    :param: output_directory: string, the output directory to save output files

    """
    def __init__(self, filename, output_directory = None, record_time = False, format ='txt'):
        self.filename = filename
        self.record_time = record_time
        self.output_directory = output_directory
        self.format = format

    def fit(self, X, main_directory='.'):
        """
        Write DataFrame to a comma-seprated values (csv) file
        :param: X: pandas DataFrame
        :param: main_directory: string, if there is any main directory for entire cheml project
        :return: nothing
        """

        if self.output_directory:
            self.output_directory = main_directory + '/' + self.output_directory
            if not os.path.exists(self.output_directory):
                os.makedirs(self.output_directory)
            if self.record_time:
                self.file_path = '%s/%s_%s.%s'%(self.output_directory, self.filename,std_datetime_str(),self.format)
            else:
                self.file_path = '%s/%s.%s' % (self.output_directory, self.filename,self.format)
        else:
            if self.record_time:
                self.file_path = '%s/%s_%s.%s'%(main_directory,self.filename,std_datetime_str(),self.format)
            else:
                self.file_path = '%s/%s.%s' %(main_directory,self.filename,self.format)
        with open(self.file_path, 'a') as file:
            file.write('%s\n' % str(X))




import numpy as np
# Todo: convert it to pandas dataframe for the consistency

class DistanceMatrix(object):
    """ (distance_matrix)
    This function finds the distance between rows of data in the input_matrix.
    The distance matrix can later be introduced to a custom kernel function.

    Parameters
    ----------
    input_matrix: numpy array
    Each element (row) of input_matrix can be a vector, matrix or list of matrices.
    These three types of representation are originally based on possible outpts
    of Coulomb_Matrix function.

    norm_type: string or integer, optional (default = 'fro')
        norm_type defines the type of norm of the difference of two rows of input
        matrix. Here is a list of available norms:

              type          norm for matrices               norm for vectors
              ----          -----------------               ----------------
              'l2'          Frobenius norm                  l2 norm for vector
              'l1'          sum(sum(abs(x-y)))              -
              'nuc'         nuclear norm                    -
              'inf'         max(sum(abs(x-y), axis=1))      max(abs(x-y))
              '-inf'        min(sum(abs(x-y), axis=1))      min(abs(x-y))
              0             -                               sum(x!=0)
              1             max(sum(abs(x-y), axis=0))      sum(abs(x)**ord)**(1./ord)
              -1            min(sum(abs(x-y), axis=0))      sum(abs(x)**ord)**(1./ord)
              2             2-norm (largest singular value) sum(abs(x)**ord)**(1./ord)
              -2            smallest singular value         sum(abs(x)**ord)**(1./ord)

                            norm for list of matrices
                            -------------------------
              'avg'         avg(sum(d(x,y[l])+d(x[l],y))))
              'max'         avg(max(d(x,y[l])+d(x[l],y))))

            * most of these norms are provided by numpy.linalg.norm
    nCores: integer, optional (default = 1)
        number of cores for multiprocessing.

    Return
    ------
    distance matrix
    """
    def __init__(self,norm_type='fro', nCores=1):
        self.norm_type = norm_type
        self.nCores = nCores
        # Todo: mpi code is required for the multiprocessing

    def transform(self,input_matrix):
        distance_matrix = []
        dim = input_matrix.shape
        if len(dim) == 2:
            # vector: Ndata=dim[0];  Nvector_elements=dim[1]
            pass
        elif len(dim) == 3:
            # matrix: Ndata=dim[0]; Nmatrix_rows=dim[1]; Nmatrix_cols=dim[2]
            for i in range(dim[0]):
                vect = []
                for k in range(0, i):
                    vect.append(distance_matrix[k][i])
                for j in range(i, dim[0]):
                    if i == j:
                        vect.append(0.0)
                    else:
                        if self.norm_type in ['fro', 'nuc', 2, 1, -1, -2]:
                            vect.append(np.linalg.norm(input_matrix[i] - input_matrix[j], ord=self.norm_type))
                        elif self.norm_type == 'inf':
                            vect.append(np.linalg.norm(input_matrix[i] - input_matrix[j], ord=np.inf))
                        elif self.norm_type == '-inf':
                            vect.append(np.linalg.norm(input_matrix[i] - input_matrix[j], ord=-np.inf))
                        elif self.norm_type == 'abs':
                            vect.append(sum(sum(abs(input_matrix[i] - input_matrix[j]))))
                        else:
                            msg = "The norm_type '%s' is not a defined type of norm" % self.norm_type
                            raise ValueError(msg)
                distance_matrix.append(vect)
            return np.array(distance_matrix)
        elif len(dim) == 4:
            # list of matrices: Ndata=dim[0]; Nmatrices_per_row=dim[1]; Nmatrix_rows=dim[2]; Nmatrix_cols=dim[3]
            pass
        else:
            msg = "the structure of input_matrix is not supported"
            raise ValueError(msg)


def tutorial1():
    script = """
                ## (Enter,datasets)
                    << host = cheml
                    << function = load_cep_homo
                    << return_X_y = True
                    >> smiles 0

                ## (Store,file)
                    << host = cheml
                    << function = SaveFile
                    << format = smi
                    << header = False
                    << filename = smiles
                    >> 0 df
                    >> filepath 1

                ## (Prepare,feature representation)
                    << host = cheml
                    << function = RDKitFingerprint
                    << molfile = @molfile
                    >> 1 molfile

            """
    return script.strip().split('\n')


def tutorial2():
    script = """
                ## (Enter,datasets)
                    << host = cheml
                    << function = load_cep_homo
                    << return_X_y = True
                    >> smiles 0

                ## (Store,file)
                    << host = cheml
                    << function = SaveFile
                    << format = smi
                    << header = False
                    << filename = smiles
                    >> 0 df
                    >> filepath 1

                ## (Prepare,feature representation)
                    << host = cheml
                    << function = RDKitFingerprint
                    << molfile = @molfile
                    >> 1 molfile

            """
    return script.strip().split('\n')



from numpy.linalg import norm
import numpy as np
from ..voronoi.VoronoiEdge import VoronoiEdge
from ..voronoi.VoronoiVertex import VoronoiVertex
from ...geometry.Plane import Plane

class VoronoiFace:
    """Class to represent a face on a Voronoi diagram.

    Attributes
    ----------
    inside_atom : Atom
            Atom on "inside" of this face.
    outside_atom : AtomImage
        Image of atom "outside" this face.
    face_plane : Plane
        Plane of face.
    intersecting_face : Plane
        Plane defining the other face defining this edge.
    face_normal : array-like
        Normal of face.
    edges : list
        Edges defining this face.
    vertices : list
        Vertices associated with this face.
    face_distance : float
        Distance from cell center to face.
    face_center : array-like
        Center of face.
    face_area : float
        Face area. Cached result.
    tol : float
        Tolerance.
    """

    def __init__(self, inside_atom, outside_atom, radical):
        """Function to initialize a Voronoi face.

        Parameters
        ----------
        inside_atom : Atom
            Atom on "inside" of this face.
        outside_atom : AtomImage
            Image of atom "outside" this face.
        radical : bool
            Whether this face is in a radical plane tessellation.
        """

        # Atom on "inside" of this face.
        self.inside_atom = inside_atom

        # Image of atom "outside" this face.
        self.outside_atom = outside_atom

        # Plane of face.
        self.face_plane = None

        # Normal of face.
        self.face_normal = None

        # Edges defining this face.
        self.edges = []

        # Vertices associated with this face.
        self.vertices = []

        # Distance from cell center to face.
        self.face_distance = None

        # Center of the face.
        self.face_center = None

        # Face area. Cached result.
        self.face_area = np.nan

        # Tolerance.
        self.tol = 1e-8

        inside_pos = self.inside_atom.get_position_cartesian()
        outside_pos = self.outside_atom.get_position()
        diff = inside_pos - outside_pos
        atom_dist = norm(diff)

        if radical:
            self.face_distance = self.get_plane_distance(
                self.inside_atom.get_radius(), self.outside_atom.get_atom(
                ).get_radius(), atom_dist)
        else:
            self.face_distance = atom_dist / 2

        self.face_center = inside_pos - diff * self.face_distance / atom_dist

    def get_plane_distance(self, r1, r2, d):
        """Function to compute distance from the center to the closest-point on
        a radical plane dividing two atoms.

        Parameters
        ----------
        r1 : float
            Radius of atom #1.
        r2 : float
            Radius of atom #2.
        d : float
            Distance between atoms.

        Returns
        -------
        value : float
            Distance from atom #1 center to plane.

        """
        return (r1**2 - r2**2 + d**2)/ 2 / d

    def __eq__(self, other):
        """Function to check instance is equal to another face.

        Parameters
        ----------
        other : VoronoiFace
            Other face.

        Returns
        -------
        value : bool
            True if equal, else False.
        """
        if isinstance(other, VoronoiFace):
            return other.inside_atom.get_id() == self.inside_atom.get_id() \
                   and other.outside_atom.__eq__(self.outside_atom)
        return False

    def __hash__(self):
        """Function to compute hash value of the instance.

        Returns
        -------
        value : int
            Hash value.
        """
        h = 7
        h = 19 + h * id(self.inside_atom)
        h = 19 + h * id(self.outside_atom)
        return h

    def __cmp__(self, other):
        """Function to compare instance with another face.

        Parameters
        ----------
        other : VoronoiFace
            Other face.

        Returns
        -------
        value : int
            -1 if instance < other, +1 if instance > other, else 0.
        """
        if self.inside_atom.get_id() == other.inside_atom.get_id():
            return self.outside_atom.__cmp__(other.outside_atom)
        else:
            return self.inside_atom.get_id() - other.inside_atom.get_id()

    def get_inside_atom(self):
        """Function to get inside atom.

        Returns
        -------
        inside_atom : Atom
            Inside atom.
        """
        return self.inside_atom

    def get_face_center(self):
        """Function to get face center.

        Returns
        -------
        face_center : array-like
            Face center.

        """
        return self.face_center

    def get_outside_atom(self):
        """Function to get outside atom image.

        Returns
        -------
        outside_atom : AtomImage
            Outside atom image.
        """
        return self.outside_atom

    def get_plane(self):
        """Function to get plane on which this face lies.

        Returns
        -------
        face_plane : Plane
            Face plane.
        """
        if self.face_plane is None:
            inside_pos = self.inside_atom.get_position_cartesian()
            outside_pos = self.outside_atom.get_position()
            self.face_plane = Plane(p=self.face_center, normal=
                                    outside_pos-inside_pos, tolerance=1e-8)
        return self.face_plane

    def get_normal(self):
        """Function to get the normal to this face.

        Returns
        -------
        face_normal : array-like
            Face normal.
        """
        if self.face_normal is None:
            self.face_normal = self.get_plane().get_normal()
        return self.face_normal

    def n_edges(self):
        """Function to get the number of edges this face has.

        Returns
        -------
        value : int
            Number of edges.

        """
        return len(self.edges)

    def get_area(self):
        """Function to determine the surface area of this face.

        Returns
        -------
        surface_area : float
            Surface area.

        """
        # If needed compute area.
        if not np.isinf(self.face_area):
            # Get centroid of face.
            centroid = self.get_centroid()
            # Loop over all edges.
            area = 0.0
            l = len(self.vertices)
            for i in range(l):
                this_vertex = self.vertices[i]
                next_vertex = self.vertices[(i + 1) % l]
                a = this_vertex.get_position() - centroid
                b = next_vertex.get_position() - centroid

                area += norm(np.cross(a, b))


            self.face_area = area / 2

        return self.face_area

    def get_centroid(self):
        """Function to get the centroid of this face.

        Returns
        -------
        centroid : array-like
            Coordinates of the centroid.

        """
        return VoronoiVertex.get_centroid(self.vertices)

    def get_face_distance(self):
        """Function to get distance from center of the cell to this face.

        Returns
        -------
        face_distance : float
            Distance.

        """
        return self.face_distance

    def get_neighbor_distance(self):
        """Function to get distance between central atom and neighbor associated
        with this face.

        Returns
        -------
        neighbor_distance : float
            Distance in Cartesian units.
        """

        return norm(self.inside_atom.get_position_cartesian() -
                    self.outside_atom.get_position())

    def get_vertices(self):
        """Function to get the list of vertices describing this face.

        Returns
        -------
        value : list
            List of vertices.

        """
        return list(self.vertices)

    def n_vertices(self):
        """Function to get the number of vertices this face has.

        Returns
        -------
        value : int
            Number of vertices.
        """
        return len(self.vertices)

    def get_edges(self):
        """Function to get the list of edges describing this face.

        Returns
        -------
        value : list
            List of edges.
        """
        return list(self.edges)

    def get_neighboring_faces(self):
        """Function to get all faces that share a vertex with this face.

        Returns
        -------
        value : list
            List of neighboring faces.
        """
        output = []
        for e in self.edges:
            f = e.get_intersecting_face()
            if f not in output:
                output.append(f)

        return output

    def get_common_vertices(self, other_face):
        """Function to get number of common vertices between two faces.

        Parameters
        ----------
        other_face : VoronoiFace
            Other face.

        Returns
        -------
        output : set
            Set of common vertices.

        """
        output = set(self.vertices)
        return output.intersection(other_face.vertices)

    def position_relative_to_face(self, point):
        """Function to get position of point with respect to face.

        Parameters
        ----------
        point : array-like
            Point to check.

        Returns
        -------
        output : int
            1 if inside of face (i.e. closer to cell center), 0 if on
            the face, 1 if outside of the face.

        """
        plane = self.get_plane()
        if plane.contains(point):
            return 0
        offset = plane.get_offset(point=point)
        if offset > 0:
            return 1
        else:
            return -1

    def __str__(self):
        """Function to print the string representation of this face.

        Returns
        -------
        output : str
            String representation of this face.
        """
        output = str(self.inside_atom.get_id()+"->"+self.outside_atom.__str__())
        return output

    def assemble_face_from_faces(self, faces):
        """Function to construct the edges, given a list of faces that will
        intersect this face.

        Used to construct faces for the direct polyhedron only! It assumes
        that face_center is inside the center of the face.

        Parameters
        ----------
        faces : array-like
            Other faces.

        Returns
        -------
        output : bool
            Whether the face assembled correctly.

        Raises
        ------
        Exception
            If there are not enough edges to build faces.

        """

        # Generate a local copy of this list.
        other_faces = list(faces)
        if self in other_faces:
            other_faces.remove(self)

        # Find all edges.
        available_edges = []
        for other_face in other_faces:
            try:
                edge = VoronoiEdge(edge_face=self,
                                   intersecting_face=other_face)
            except Exception:
                continue

            # Store the data.
            available_edges.append(edge)

        # Error check.
        if len(available_edges) < 3:
            raise Exception("Not enough edges.")

        available_edges.sort(cmp=self.edge_comp)
        return self.assemble_face_from_edges(available_edges)

    def edge_comp(self, a, b):
        """Function to compare two edges.

        Parameters
        ----------
        a : VoronoiEdge
            Edge 1.
        b : VoronoiEdge
            Edge 2.

        Returns
        -------
        output : int
            1 if a < b, +1 if a > b, else 0.

        """
        d1 = a.get_line().distance(p=self.face_center)
        d2 = b.get_line().distance(p=self.face_center)

        # Have to check if they are really close. If they are, return 0.
        if (d1 - d2) ** 2 < 1e-30:
            return 0
        elif d1 < d2:
            return -1
        else:
            return +1

    def assemble_face_from_edges(self, available_edges):
        """Function to assemble face, given a list of edges.

        Parameters
        ----------
        available_edges : array-like
            List of edges.

        Returns
        -------
        output : bool
            Whether face is closed.

        Raises
        ------
        Exception
            If face failed to build.
        """

        # Clear cached face area.
        self.face_area = np.nan

        # The closest edge is certainly on the face.
        face_edges = []
        face_edges.append(available_edges[0])
        cur_edge = face_edges[0]

        # Now, loop through all the edges and compute their intersection with
        # all other edges on the face. This is accomplished by finding the
        # next edge and proceeding until we reach the beginning or none is
        # found.
        while True:
            next_edge = cur_edge.find_next_edge(available_edges)
            if next_edge == face_edges[0]:
                break
            elif next_edge is None:
                raise Exception("Face is not closed.")
            elif len(face_edges) > len(available_edges):
                # This happens if the first edge is not a valid edge.
                # find the first edge and continues from a different starting
                # point. To address this, easy way is to remove the problem
                # edge and try again.
                new_avail = list(available_edges)
                new_avail.remove(new_avail[0])
                if len(new_avail) < 3:
                    raise Exception("Face failed to build.")
                return self.assemble_face_from_edges(new_avail)
            face_edges.append(next_edge)
            cur_edge = next_edge

        # Compute intersections and add edges.
        self.edges = []
        l_fe = len(face_edges)
        for i in range(l_fe):
            VoronoiEdge.compute_intersection(face_edges[i], face_edges[(i +
                                    1) % l_fe], just_join=True)
            self.edges.append(face_edges[i])

        # Store vertices.
        self.compute_vertices()

        return self.is_closed()


    def compute_vertices(self):
        """Function to recompute vertices. Should be run after each time the
        edges are updated.

        Raises
        ------
        Exception
            If an error occurs while computing vertex.
        """
        self.vertices = []
        for e in self.edges:
            try:
                self.vertices.append(e.get_start_vertex())
            except Exception:
                raise Exception("Vertex computation error.")

        # Compute area.
        self.get_area()

    def is_closed(self):
        """Function to check if face is closed.

        Returns
        -------
        output : bool
            True if face is closed, else False.
        """
        if len(self.edges) < 3:
            return False

        for i in range(len(self.vertices)-1):
            if not (self.edges[i].get_next_edge() in self.edges and
                        self.edges[i].get_previous_edge() in self.edges):
                return False
        return True

    def compute_intersection(self, new_face):
        """Function to compute intersection between this face and a new face.

        If the face intersects, update the list of edges.
        Special case: If a current edge of this face is completely on the new
        face (i.e., both vertices are on the face), that edge will be
        replaced with a new edge that is the result of the intersection
        between the new face and this face.

        Parameters
        ----------
        new_face : VoronoiFace
            Face in consideration.

        Returns
        -------
        output : VoronoiEdge
            If the new face intersect, return the pair of the edge that was
            formed. Returns None if no edge was formed.

        Raises
        ------
        Exception
            If new face does not intersect this face.
            If face was not closed to begin with.
            If edge on the new face was not found.
            If new face doesn't intersect this face.
            If edge not found.
            If new edge doesn't intersect current ones.
        """

        # Clear cached area.
        self.face_area = np.nan

        # See if any vertex is outside this new face or if any vertices are
        # on this face.
        outside_vertex = None
        n_c = 0
        for v in self.vertices:
            rel_pos = new_face.position_relative_to_face(v.get_position())
            if rel_pos > 0:
                outside_vertex = v
                break
            elif rel_pos == 0:
                n_c += 1

        # If it doesn't intersect, return "None".
        if outside_vertex is None:
            # Check if two vertices are on the new face. If so, they must
            # share an edge because these faces are convex.
            if n_c < 2:
                return None

            # If so, find what edge they correspond to.
            new_edge = None
            edge_to_replace = None
            try:
                new_edge = VoronoiEdge(self, new_face)
            except Exception:
                raise Exception("New face does not intersect this face.")

            try:
                for e in self.edges:
                    if new_face.position_relative_to_face(e.get_start_vertex(
                    ).get_position()) == 0 and \
                    new_face.position_relative_to_face(
                    e.get_end_vertex().get_position()) == 0:
                        edge_to_replace = e
                        break
            except Exception:
                raise Exception("Face was not closed to begin with.")

            # Compute the new edges.
            if edge_to_replace is None:
                raise Exception("Edge on the new face was not found.")

            try:
                VoronoiEdge.compute_intersection(new_edge,
                    edge_to_replace.get_previous_edge(), just_join=True)
                VoronoiEdge.compute_intersection(new_edge,
                    edge_to_replace.get_next_edge(), just_join=True)
            except Exception:
                raise Exception("New edge and old are not parallel.")

            # Replace edge in the lineup.
            self.edges[self.edges.index(edge_to_replace)] = new_edge

            # Recompute vertices.
            self.compute_vertices()

            return new_edge.generate_pair()

        # Get this vertex ID number.
        outside_vertex_id = self.vertices.index(outside_vertex)

        # Traverse forward until a vertex inside is found.
        last_vertex = outside_vertex_id
        l_v = len(self.vertices)
        while new_face.position_relative_to_face(self.vertices[
                                    last_vertex].get_position()) >= 0:
            last_vertex = (last_vertex + 1) %  l_v
            # If we go all the way around the loop, all vertices are outside
            # this face. Clear everything and return "None" because no edge
            # is formed.
            if outside_vertex_id == last_vertex:
                self.edges = []
                self.vertices = []
                return None

        # Traverse backwards until you find the first vertex that is inside
        # the intersecting plane.
        first_vertex = outside_vertex_id
        while new_face.position_relative_to_face(self.vertices[
                                    first_vertex].get_position()) >= 0:
            first_vertex -= 1
            if first_vertex < 0:
                first_vertex += l_v

        # Create the new edge.
        new_edge = None
        try:
            new_edge = VoronoiEdge(self, new_face)
        except Exception:
            raise Exception("New face doesn't intersect this face.")

        start_edge = self.vertices[first_vertex].get_next_edge()
        end_edge = self.vertices[last_vertex].get_previous_edge()
        # Eliminate edges that are outside the face.
        cur_edge = start_edge.get_next_edge()
        while cur_edge != end_edge:
            if cur_edge not in self.edges:
                raise Exception("Edge not found.")
            else:
                self.edges.remove(cur_edge)

            cur_edge = cur_edge.get_next_edge()

        # Join start and edges with this new edge.
        try:
            VoronoiEdge.compute_intersection(start_edge, new_edge,
                                             just_join=True)
            VoronoiEdge.compute_intersection(end_edge, new_edge, just_join=True)
        except Exception:
            raise Exception("New edge doesn't intersect current ones.")

        # Add it to the edges list.
        idx = self.edges.index(start_edge)
        self.edges.insert(1 + idx, new_edge)
        # Recompute vertices.
        self.compute_vertices()
        return new_edge.generate_pair()

    def reset_edges(self, edges):
        """Function to reset edges to a previous state. Used in case cell
        intersection fails.

        Parameters
        ----------
        edges : array-like
            Previous list of edges.

        Raises
        ------
        RuntimeError
            If the stored state was in error.

        """
        try:
            l_e = len(edges)
            for i in range(l_e):
                VoronoiEdge.compute_intersection(edges[i], edges[ (i + 1) %
                                            l_e], just_join=True)
            self.edges = edges
            self.compute_vertices()
        except Exception:
            raise RuntimeError("Was the stored state in error?")


    def get_cut_length(self, new_face):
        """Function to compute length of new edge created by intersection.

        Parameters
        ----------
        new_face : VoronoiFace
            Face intersecting with this one.

        Returns
        -------
        output : float
            Cut length.

        Raises
        ------
        Exception
            New face does not intersect this face.
        """

        # Find a single vertex that is outside the face.
        outside_vertex = None
        for v in self.vertices:
            if new_face.position_relative_to_face(v.get_position()) > 0:
                outside_vertex = v
                break

        # If no vertex is outside, cut length == 0.
        if outside_vertex is None:
            return 0

        # Get the vertex ID number.
        outside_vertex_id = self.vertices.index(outside_vertex)

        # Traverse forward until a vertex inside is found.
        last_vertex = outside_vertex_id
        l_v = len(self.vertices)
        while new_face.position_relative_to_face(self.vertices[
                                                     last_vertex].get_position()) >= 0:
            last_vertex = (last_vertex + 1) % l_v
            # If we go all the way around the loop, all vertices are outside
            # this face. Clear everything and return "None" because no edge
            # is formed.
            if outside_vertex_id == last_vertex:
                self.edges = []
                self.vertices = []
                return float("inf")

        # Traverse backwards until you find the first vertex that is inside
        # the intersecting plane.
        first_vertex = outside_vertex_id
        while new_face.position_relative_to_face(self.vertices[
                                                     first_vertex].get_position()) >= 0:
            first_vertex -= 1
            if first_vertex < 0:
                first_vertex += l_v

        # Create the new edge.
        new_edge = None
        try:
            new_edge = VoronoiEdge(self, new_face)
        except Exception:
            raise Exception("New face doesn't intersect this face.")

        start_edge = self.vertices[first_vertex].get_next_edge()
        end_edge = self.vertices[last_vertex].get_previous_edge()

        p1 = start_edge.get_line().intersection(new_edge.get_line())
        p2 = end_edge.get_line().intersection(new_edge.get_line())
        return norm(p1 - p2)


    def is_contacted_by(self, other_face):
        """Function to check whether instance if contacted by another face.

        This occurs if any of the vertices are outside or on the other face.

        Parameters
        ----------
        other_face : VoronoiFace
            Other face.

        Returns
        -------
        output : bool
            Whether it contacts instance.

        """
        for v in self.vertices:
            if other_face.position_relative_to_face(v.get_position()) >= 0:
                return True
        return False

    def is_completely_outside(self, other_face):
        """Function to check whether instance is completely outside of another
        face.

        Parameters
        ----------
        other_face : VoronoiFace
            Other face.

        Returns
        -------
        output : bool
            Whether any vertex of instance is in outside or in contact
            with the other face.

        """
        for v in self.vertices:
            if other_face.position_relative_to_face(v.get_position()) < 0:
                return False
        return True

    def contains_vertex(self, v):
        """Function to check instance contains a vertex.

        Parameters
        ----------
        v : VoronoiVertex
            Vertex to check.

        Returns
        -------
        output : bool
            True if instance contains v, else False.

        """
        return v in self.vertices


from setuptools import setup, find_packages
from os import path
import cheml

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md')) as f:
    long_description = f.read()

setup(  name = 'chemml',
        version = cheml.__version__,
        author='Mojtaba Haghighatlari, Johannes Hachmann',
        author_email='mojtabah@buffalo.edu, hachmann@buffalo.edu',
        # url='https://github.com/hachmannlab/chemml',
        project_urls={
            'Source': 'https://github.com/hachmannlab/chemml',
            'url': 'https://hachmannlab.github.io/chemml/'
        },
        description='A Machine Learning and Informatics Program Suite for the Chemical and Materials Sciences',
        long_description=long_description,

        packages = find_packages(exclude=['docs']),

        scripts = ['bin/chemmlwrapper'],

        keywords=['Machine Learning', 'Data Mining', 'Quantum Chemistry', 'Materials Science', 'Informatics'],
        license='3-Clause BSD',
        classifiers=['Development Status :: 4 - Beta',
                   'Natural Language :: English',
                   'Programming Language :: Python :: 2.7',
                  ],

        python_requires = '>=2.7, <3',

        install_requires = ['numpy>=1.13', 'pandas>=0.20.3', 'tensorflow==1.1.0', 'keras==2.1.5',
                            'scikit-learn==0.19.1', 'babel>=2.3.4', 'matplotlib>=1.5.1', 'deap>=1.2.2',
                            'lxml','nose','ipywidgets>=7.1','widgetsnbextension>=3.1','graphviz'],

        test_suite='nose.collector',
        tests_require=['nose'],

        include_package_data = True,
        # package_data={
        #                 '': ['*.xyz', '*.csv', '*.vasp', '*.txt'],
        #                 # 'cheml': ['datasets/data/*', 'tests/data/*', 'tests/configfiles/*'],
        #             },
        extras_require={
                'gui': ['ipywidgets', 'graphviz'],
                'ml': ['scikit-learn>=0.18.1','tensorflow>=1.4.1', 'keras>=2.1.2'],
                'viz': ['matplotlib>=1.5.1'],
                'ga': ['deap>=1.2.2'],

            },
     )
